import datetime
import json
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from uaclient import exceptions, messages, util

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class IncorrectTypeError(exceptions.UbuntuProError):
    _formatted_msg = messages.E_INCORRECT_TYPE_ERROR_MESSAGE
    expected_type = None  # type: str
    got_type = None  # type: str


class IncorrectListElementTypeError(IncorrectTypeError):
    _formatted_msg = messages.E_INCORRECT_LIST_ELEMENT_TYPE_ERROR_MESSAGE

    def __init__(self, *, err: IncorrectTypeError, at_index: int):
        super().__init__(index=at_index, nested_msg=err.msg)
        self.expected_type = err.expected_type
        self.got_type = err.got_type


class IncorrectDictElementTypeError(IncorrectTypeError):
    _formatted_msg = messages.E_INCORRECT_DICT_ELEMENT_TYPE_ERROR_MESSAGE

    def __init__(self, *, err: IncorrectTypeError, key: str, value: Any):
        super().__init__(
            key=key, value_type=type(value).__name__, nested_msg=err.msg
        )
        self.expected_type = err.expected_type
        self.got_type = err.got_type


class IncorrectFieldTypeError(IncorrectTypeError):
    _formatted_msg = messages.E_INCORRECT_FIELD_TYPE_ERROR_MESSAGE
    key = None  # type: str

    def __init__(self, *, err: IncorrectTypeError, key: str):
        super().__init__(key=key, nested_msg=err.msg)
        self.expected_type = err.expected_type
        self.got_type = err.got_type


class IncorrectEnumValueError(IncorrectTypeError):
    _formatted_msg = messages.E_INCORRECT_ENUM_VALUE_ERROR_MESSAGE

    def __init__(self, *, values: List[Union[str, int]], enum_class: Any):
        super().__init__(values=values, enum_class=enum_class.__name__)
        self.expected_type = "one of: {}".format(
            ", ".join([str(v) for v in values])
        )
        self.got_type = "<invalid_value>"


class DataValue:
    """
    Generic data value to be extended by more specific typed data values.
    This establishes the interface of a static/class method called `from_value`
    that returns the parsed value if appropriate.
    """

    @staticmethod
    def from_value(val: Any) -> Any:
        return val


E = TypeVar("E", bound="EnumDataValue")


class EnumDataValue(DataValue, Enum):
    """
    To be used for parsing enum values
    from_value raises an error if the value is not in the enum class values
    and returns the value if found.
    """

    @classmethod
    def from_value(cls: Type[E], val: Any) -> E:
        try:
            return cls(val)
        except ValueError:
            values = [i.value for i in cls]
            raise IncorrectEnumValueError(values=values, enum_class=cls)


class StringDataValue(DataValue):
    """
    To be used for parsing string values
    from_value raises an error if the value is not a string and returns
    the string itself if it is a string.
    """

    python_type_name = "str"

    @staticmethod
    def from_value(val: Any) -> str:
        if not isinstance(val, str):
            raise IncorrectTypeError(
                expected_type="str", got_type=type(val).__name__
            )
        return val


class IntDataValue(DataValue):
    """
    To be used for parsing int values
    from_value raises an error if the value is not a int and returns
    the int itself if it is a int.
    """

    python_type_name = "int"

    @staticmethod
    def from_value(val: Any) -> int:
        if not isinstance(val, int) or isinstance(val, bool):
            raise IncorrectTypeError(
                expected_type="int", got_type=type(val).__name__
            )
        return val


class FloatDataValue(DataValue):
    """
    To be used for parsing float values
    from_value raises an error if the value is not a float and returns
    the float itself if it is a float.
    """

    python_type_name = "float"

    @staticmethod
    def from_value(val: Any) -> float:
        if not isinstance(val, float) and not isinstance(val, int):
            raise IncorrectTypeError(
                expected_type="float", got_type=type(val).__name__
            )
        return float(val)


class BoolDataValue(DataValue):
    """
    To be used for parsing bool values
    from_value raises an error if the value is not a bool and returns
    the bool itself if it is a bool.
    """

    python_type_name = "bool"

    @staticmethod
    def from_value(val: Any) -> bool:
        if not isinstance(val, bool):
            raise IncorrectTypeError(
                expected_type="bool", got_type=type(val).__name__
            )
        return val


class DatetimeDataValue(DataValue):
    """
    This expects that value is a datetime.
    from_value raises an error if the value is not a datetime and returns
    the datetime itself if it is a datetime.
    """

    python_type_name = "datetime"

    @staticmethod
    def from_value(val: Any) -> datetime.datetime:
        if not isinstance(val, datetime.datetime):
            raise IncorrectTypeError(
                expected_type="datetime", got_type=type(val).__name__
            )
        return val


def data_list(data_cls: Type[DataValue]) -> Type[DataValue]:
    """
    To be used for parsing lists of a certain DataValue type.
    Returns a class that extends DataValue and validates that
    each item in a list is the correct type in its from_value.
    """

    class _DataList(DataValue):
        item_cls = data_cls

        @staticmethod
        def from_value(val: Any) -> List:
            if not isinstance(val, list):
                raise IncorrectTypeError(
                    expected_type="list", got_type=type(val).__name__
                )
            new_val = []
            for i, item in enumerate(val):
                try:
                    new_val.append(data_cls.from_value(item))
                except IncorrectTypeError as e:
                    raise IncorrectListElementTypeError(err=e, at_index=i)
            return new_val

    return _DataList


def data_dict(value_cls: Type[DataValue]) -> Type[DataValue]:
    """
    To be used for parsing dicts of a certain DataValue type.
    Returns a class that extends DataValue and validates that
    each key, value is the correct type in its from_value.
    """

    class _DataDict(DataValue):
        python_type_name = "dict"
        dict_value_cls = value_cls

        @staticmethod
        def from_value(val: Any) -> Dict:
            if not isinstance(val, dict):
                raise IncorrectTypeError(
                    expected_type="dict", got_type=type(val).__name__
                )
            new_val = {}
            for key, value in val.items():
                try:
                    new_val[StringDataValue.from_value(key)] = (
                        value_cls.from_value(value)
                    )
                except IncorrectTypeError as e:
                    raise IncorrectDictElementTypeError(
                        err=e, key=key, value=value
                    )
            return new_val

    return _DataDict


def data_list_to_list(
    val: List[Union["DataObject", list, str, int, bool, Enum]],
    keep_none: bool = True,
) -> list:
    new_val = []  # type: list
    for item in val:
        if isinstance(item, DataObject):
            new_val.append(item.to_dict(keep_none))
        elif isinstance(item, list):
            new_val.append(data_list_to_list(item, keep_none))
        elif isinstance(item, dict):
            new_val.append(data_dict_to_dict(item, keep_none))
        elif isinstance(item, Enum):
            new_val.append(item.value)
        else:
            new_val.append(item)
    return new_val


def data_dict_to_dict(
    val: Dict["DataValue", "DataValue"],
    keep_none: bool = True,
) -> dict:
    new_val = {}  # type: Dict[Any, Any]
    for key, value in val.items():
        if isinstance(value, DataObject):
            new_val[key] = value.to_dict(keep_none)
        elif isinstance(value, dict):
            new_val[key] = data_dict_to_dict(value, keep_none)
        elif isinstance(value, list):
            new_val[key] = data_list_to_list(value, keep_none)
        elif isinstance(value, Enum):
            new_val[key] = value.value
        else:
            new_val[key] = value
    return new_val


class Field:
    """
    For defining the fields static property of a DataObject.
    """

    def __init__(
        self,
        key: str,
        data_cls: Type[DataValue],
        required: bool = True,
        dict_key: Optional[str] = None,
        doc: Optional[str] = None,
    ):
        self.key = key
        self.data_cls = data_cls
        self.required = required
        if dict_key is not None:
            self.dict_key = dict_key
        else:
            self.dict_key = self.key
        self.doc = doc


T = TypeVar("T", bound="DataObject")


class DataObject(DataValue):
    """
    For defining a python object that can be parsed from a dict.
    Validates that a set of expected fields are present in the dict
    that is parsed and that the values of those fields are the correct
    DataValue by calling from_value on each.
    The fields are defined using the `fields` static property.
    DataObjects can be used in Fields of other DataObjects.
    To define a new DataObject:
      1. Create a new class that extends DataObject.
      2. Define the `fields` static property to be a list of Field objects
      3. Define the constructor to take kwargs that match the list of Field
         objects.
           a. Example 1: Field("keyname", StringDataValue) -> keyname: str
           b. Example 2: Field("keyname", data_list(IntDataValue), required=False) -> keyname: Optional[List[int]]  # noqa: E501
      4. Use from_value or from_dict to parse a dict into the python object.
    """

    fields = []  # type: List[Field]

    def __init__(self, **_kwargs):
        pass

    def __eq__(self, other):
        for field in self.fields:
            self_val = getattr(self, field.key, None)
            other_val = getattr(other, field.key, None)
            if self_val != other_val:
                return False
        return True

    def __repr__(self):
        return "{}{}".format(
            self.__class__.__name__, self.to_dict().__repr__()
        )

    def to_dict(self, keep_none: bool = True) -> dict:
        d = {}
        for field in self.fields:
            val = getattr(self, field.key, None)
            new_val = None  # type: Any

            if isinstance(val, DataObject):
                new_val = val.to_dict(keep_none)
            elif isinstance(val, list):
                new_val = data_list_to_list(val, keep_none)
            elif isinstance(val, dict):
                new_val = data_dict_to_dict(val, keep_none)
            elif isinstance(val, Enum):
                new_val = val.value
            else:
                # simple type, just copy
                new_val = val

            if new_val is not None or keep_none:
                d[field.dict_key] = new_val
        return d

    def to_json(self, keep_null: bool = True) -> str:
        return json.dumps(
            self.to_dict(keep_none=keep_null),
            sort_keys=True,
            cls=util.DatetimeAwareJSONEncoder,
        )

    @classmethod
    def from_dict(
        cls: Type[T], d: dict, optional_type_errors_become_null: bool = False
    ) -> T:
        kwargs = {}
        for field in cls.fields:
            try:
                val = d[field.dict_key]
            except KeyError:
                if field.required:
                    raise IncorrectFieldTypeError(
                        err=IncorrectTypeError(
                            expected_type=field.data_cls.__name__,
                            got_type="null",
                        ),
                        key=field.dict_key,
                    )
                else:
                    val = None
            if val is not None:
                try:
                    val = field.data_cls.from_value(val)
                except IncorrectTypeError as e:
                    if not field.required and optional_type_errors_become_null:
                        LOG.warning(
                            "%s is wrong type (expected %s but got %s) but "
                            "considered optional - treating as null",
                            field.key,
                            e.expected_type,
                            e.got_type,
                        )
                        val = None
                    else:
                        raise IncorrectFieldTypeError(
                            err=e, key=field.dict_key
                        )

            kwargs[field.key] = val
        return cls(**kwargs)

    @classmethod
    def from_value(cls, val: Any):
        if not isinstance(val, dict):
            raise IncorrectTypeError(
                expected_type="dict", got_type=type(val).__name__
            )
        return cls.from_dict(val)


class AttachActionsConfigFile(DataObject):
    """
    The format of the yaml file that can be passed with
    pro attach --attach-config /path/to/file
    """

    fields = [
        Field("token", StringDataValue),
        Field("enable_services", data_list(StringDataValue), required=False),
    ]

    def __init__(self, *, token: str, enable_services: Optional[List[str]]):
        self.token = token
        self.enable_services = enable_services



import datetime
import json

import mock
import pytest

from uaclient import exceptions, http, messages, system
from uaclient.entitlements.livepatch import LivepatchEntitlement
from uaclient.files.state_files import LivepatchSupportCacheData
from uaclient.livepatch import (
    LIVEPATCH_CMD,
    LivepatchPatchFixStatus,
    LivepatchPatchStatus,
    LivepatchStatusStatus,
    LivepatchSupport,
    UALivepatchClient,
    _on_supported_kernel_api,
    _on_supported_kernel_cache,
    _on_supported_kernel_cli,
    configure_livepatch_proxy,
    get_config_option_value,
    on_supported_kernel,
    status,
    unconfigure_livepatch_proxy,
)

M_PATH = "uaclient.livepatch."


class TestStatus:
    @pytest.mark.parametrize(
        [
            "is_installed",
            "subp_sideeffect",
            "expected",
        ],
        [
            (False, None, None),
            (True, [("", None)], None),
            (True, [("{", None)], None),
            (True, [("{}", None)], None),
            (True, [('{"Status": false}', None)], None),
            (True, [('{"Status": []}', None)], None),
            (
                True,
                [('{"Status": [{}]}', None)],
                LivepatchStatusStatus(
                    kernel=None, livepatch=None, supported=None
                ),
            ),
            (
                True,
                [
                    (
                        json.dumps(
                            {
                                "Status": [
                                    {
                                        "Kernel": "installed-kernel-generic",
                                        "Livepatch": {
                                            "State": "nothing-to-apply",
                                            "Version": "100",
                                        },
                                    }
                                ],
                            }
                        ),
                        None,
                    )
                ],
                LivepatchStatusStatus(
                    kernel="installed-kernel-generic",
                    livepatch=LivepatchPatchStatus(
                        state="nothing-to-apply",
                        fixes=None,
                        version="100",
                    ),
                    supported=None,
                ),
            ),
            (
                True,
                [
                    (
                        json.dumps(
                            {
                                "Status": [
                                    {
                                        "Kernel": "installed-kernel-generic",
                                        "Livepatch": {
                                            "State": "applied",
                                            "Fixes": [
                                                {
                                                    "Name": "cve-example",
                                                    "Description": "",
                                                    "Bug": "",
                                                    "Patched": True,
                                                },
                                            ],
                                        },
                                    }
                                ],
                            }
                        ),
                        None,
                    )
                ],
                LivepatchStatusStatus(
                    kernel="installed-kernel-generic",
                    livepatch=LivepatchPatchStatus(
                        state="applied",
                        fixes=[
                            LivepatchPatchFixStatus(
                                name="cve-example",
                                patched=True,
                            )
                        ],
                        version=None,
                    ),
                    supported=None,
                ),
            ),
            (
                True,
                [
                    (
                        json.dumps(
                            {
                                "Client-Version": "version",
                                "Machine-Id": "machine-id",
                                "Architecture": "x86_64",
                                "CPU-Model": "Intel(R) Core(TM) i7-8650U CPU @ 1.90GHz",  # noqa: E501
                                "Last-Check": "2022-07-05T18:29:00Z",
                                "Boot-Time": "2022-07-05T18:27:12Z",
                                "Uptime": "203",
                                "Status": [
                                    {
                                        "Kernel": "4.15.0-187.198-generic",
                                        "Running": True,
                                        "Livepatch": {
                                            "CheckState": "checked",
                                            "State": "nothing-to-apply",
                                            "Version": "",
                                        },
                                    }
                                ],
                                "tier": "stable",
                            }
                        ),
                        None,
                    )
                ],
                LivepatchStatusStatus(
                    kernel="4.15.0-187.198-generic",
                    livepatch=LivepatchPatchStatus(
                        state="nothing-to-apply", fixes=None, version=""
                    ),
                    supported=None,
                ),
            ),
            (
                True,
                [
                    (
                        json.dumps(
                            {
                                "Status": [
                                    {
                                        "Supported": "supported",
                                    }
                                ],
                            }
                        ),
                        None,
                    )
                ],
                LivepatchStatusStatus(
                    kernel=None,
                    livepatch=None,
                    supported="supported",
                ),
            ),
        ],
    )
    @mock.patch(M_PATH + "system.subp")
    @mock.patch(M_PATH + "is_livepatch_installed")
    def test_status(
        self,
        m_is_livepatch_installed,
        m_subp,
        is_installed,
        subp_sideeffect,
        expected,
    ):
        m_is_livepatch_installed.return_value = is_installed
        m_subp.side_effect = subp_sideeffect
        assert expected == status()

    @mock.patch(M_PATH + "system.subp")
    @mock.patch(M_PATH + "is_livepatch_installed")
    def test_status_when_command_fails(
        self,
        m_is_livepatch_installed,
        m_subp,
    ):

        m_is_livepatch_installed.return_value = True
        m_subp.side_effect = exceptions.ProcessExecutionError("")

        with pytest.raises(exceptions.ProcessExecutionError):
            status()


@mock.patch(M_PATH + "serviceclient.UAServiceClient.request_url")
@mock.patch(M_PATH + "serviceclient.UAServiceClient.headers")
class TestUALivepatchClient:
    @pytest.mark.parametrize(
        [
            "version",
            "flavor",
            "arch",
            "codename",
            "build_date",
            "expected_request_calls",
        ],
        [
            (
                "1.23-4",
                "generic",
                "amd64",
                "xenial",
                datetime.datetime(
                    2020, 1, 1, 1, 1, 1, tzinfo=datetime.timezone.utc
                ),
                [
                    mock.call(
                        "/v1/api/kernels/supported",
                        headers=mock.ANY,
                        query_params={
                            "kernel-version": "1.23-4",
                            "flavour": "generic",
                            "architecture": "amd64",
                            "codename": "xenial",
                            "build-date": "2020-01-01T01:01:01+00:00",
                        },
                    )
                ],
            ),
            (
                "5.67-8",
                "kvm",
                "arm64",
                "kinetic",
                None,
                [
                    mock.call(
                        "/v1/api/kernels/supported",
                        headers=mock.ANY,
                        query_params={
                            "kernel-version": "5.67-8",
                            "flavour": "kvm",
                            "architecture": "arm64",
                            "codename": "kinetic",
                            "build-date": "unknown",
                        },
                    )
                ],
            ),
        ],
    )
    def test_is_kernel_supported_calls_api_with_correct_params(
        self,
        m_headers,
        m_request_url,
        version,
        flavor,
        arch,
        codename,
        build_date,
        expected_request_calls,
    ):
        lp_client = UALivepatchClient()
        lp_client.is_kernel_supported(
            version, flavor, arch, codename, build_date
        )
        assert m_request_url.call_args_list == expected_request_calls

    @pytest.mark.parametrize(
        [
            "request_side_effect",
            "expected",
        ],
        [
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="",
                        json_dict={"Supported": True},
                        json_list=[],
                    )
                ],
                LivepatchSupport.SUPPORTED,
            ),
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="",
                        json_dict={"Supported": False},
                        json_list=[],
                    )
                ],
                LivepatchSupport.UNSUPPORTED,
            ),
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="string",
                        json_dict={},
                        json_list=[],
                    )
                ],
                LivepatchSupport.UNSUPPORTED,
            ),
            (
                exceptions.ConnectivityError(
                    cause=mock.MagicMock(), url="url"
                ),
                None,
            ),
            (Exception(), None),
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="",
                        json_dict={"Supported": "supported"},
                        json_list=[],
                    )
                ],
                LivepatchSupport.SUPPORTED,
            ),
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="",
                        json_dict={"Supported": "unsupported"},
                        json_list=[],
                    )
                ],
                LivepatchSupport.UNSUPPORTED,
            ),
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="",
                        json_dict={"Supported": "unknown"},
                        json_list=[],
                    )
                ],
                LivepatchSupport.UNKNOWN,
            ),
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="",
                        json_dict={"Supported": "kernel-end-of-life"},
                        json_list=[],
                    )
                ],
                LivepatchSupport.KERNEL_EOL,
            ),
            (
                [
                    http.HTTPResponse(
                        code=200,
                        headers={},
                        body="",
                        json_dict={"Supported": "kernel-upgrade-required"},
                        json_list=[],
                    )
                ],
                LivepatchSupport.KERNEL_UPGRADE_REQUIRED,
            ),
        ],
    )
    def test_is_kernel_supported_interprets_api_response(
        self,
        m_headers,
        m_request_url,
        request_side_effect,
        expected,
    ):
        m_request_url.side_effect = request_side_effect
        lp_client = UALivepatchClient()
        assert lp_client.is_kernel_supported("", "", "", "", None) == expected


class TestOnSupportedKernel:
    @pytest.mark.parametrize(
        [
            "livepatch_status",
            "expected",
        ],
        [
            (None, None),
            (
                LivepatchStatusStatus(
                    kernel=None, livepatch=None, supported=None
                ),
                None,
            ),
            (
                LivepatchStatusStatus(
                    kernel=None, livepatch=None, supported="supported"
                ),
                LivepatchSupport.SUPPORTED,
            ),
            (
                LivepatchStatusStatus(
                    kernel=None, livepatch=None, supported="unsupported"
                ),
                LivepatchSupport.UNSUPPORTED,
            ),
            (
                LivepatchStatusStatus(
                    kernel=None,
                    livepatch=None,
                    supported="kernel-upgrade-required",
                ),
                LivepatchSupport.KERNEL_UPGRADE_REQUIRED,
            ),
            (
                LivepatchStatusStatus(
                    kernel=None, livepatch=None, supported="kernel-end-of-life"
                ),
                LivepatchSupport.KERNEL_EOL,
            ),
            (
                LivepatchStatusStatus(
                    kernel=None, livepatch=None, supported="unknown"
                ),
                LivepatchSupport.UNKNOWN,
            ),
        ],
    )
    @mock.patch(M_PATH + "status")
    def test_on_supported_kernel_cli(
        self,
        m_livepatch_status,
        livepatch_status,
        expected,
    ):
        m_livepatch_status.return_value = livepatch_status
        assert _on_supported_kernel_cli() == expected

    @pytest.mark.parametrize(
        [
            "args",
            "cache_contents",
            "expected",
        ],
        [
            # valid true
            (
                ("5.14-14", "generic", "amd64", "focal"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=6),
                    supported=True,
                ),
                (True, True),
            ),
            # valid false
            (
                ("5.14-14", "generic", "amd64", "focal"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=6),
                    supported=False,
                ),
                (True, False),
            ),
            # valid none
            (
                ("5.14-14", "generic", "amd64", "focal"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=6),
                    supported=None,
                ),
                (True, None),
            ),
            # invalid version doesn't match
            (
                ("5.14-13", "generic", "amd64", "focal"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=6),
                    supported=True,
                ),
                (False, None),
            ),
            # invalid flavor doesn't match
            (
                ("5.14-14", "kvm", "amd64", "focal"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=6),
                    supported=True,
                ),
                (False, None),
            ),
            # invalid arch doesn't match
            (
                ("5.14-14", "generic", "arm64", "focal"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=6),
                    supported=True,
                ),
                (False, None),
            ),
            # invalid codename doesn't match
            (
                ("5.14-14", "generic", "amd64", "xenial"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=6),
                    supported=True,
                ),
                (False, None),
            ),
            # invalid too old
            (
                ("5.14-14", "generic", "amd64", "xenial"),
                LivepatchSupportCacheData(
                    version="5.14-14",
                    flavor="generic",
                    arch="amd64",
                    codename="focal",
                    cached_at=datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=8),
                    supported=True,
                ),
                (False, None),
            ),
        ],
    )
    @mock.patch(M_PATH + "state_files.livepatch_support_cache.read")
    def test_on_supported_kernel_cache(
        self,
        m_cache_read,
        args,
        cache_contents,
        expected,
    ):
        m_cache_read.return_value = cache_contents
        assert _on_supported_kernel_cache(*args) == expected

    @pytest.mark.parametrize(
        [
            "args",
            "is_kernel_supported",
            "expected_cache_write_call_args",
            "expected",
        ],
        [
            (
                ("5.14-14", "generic", "amd64", "focal", None),
                LivepatchSupport.SUPPORTED,
                [
                    mock.call(
                        LivepatchSupportCacheData(
                            version="5.14-14",
                            flavor="generic",
                            arch="amd64",
                            codename="focal",
                            supported=True,
                            cached_at=mock.ANY,
                        )
                    )
                ],
                LivepatchSupport.SUPPORTED,
            ),
            (
                ("5.14-14", "kvm", "arm64", "focal", None),
                LivepatchSupport.UNSUPPORTED,
                [
                    mock.call(
                        LivepatchSupportCacheData(
                            version="5.14-14",
                            flavor="kvm",
                            arch="arm64",
                            codename="focal",
                            supported=False,
                            cached_at=mock.ANY,
                        )
                    )
                ],
                LivepatchSupport.UNSUPPORTED,
            ),
            (
                ("4.14-14", "kvm", "arm64", "xenial", None),
                None,
                [
                    mock.call(
                        LivepatchSupportCacheData(
                            version="4.14-14",
                            flavor="kvm",
                            arch="arm64",
                            codename="xenial",
                            supported=None,
                            cached_at=mock.ANY,
                        )
                    )
                ],
                None,
            ),
            (
                ("4.14-14", "kvm", "arm64", "xenial", None),
                LivepatchSupport.UNKNOWN,
                [
                    mock.call(
                        LivepatchSupportCacheData(
                            version="4.14-14",
                            flavor="kvm",
                            arch="arm64",
                            codename="xenial",
                            supported=None,
                            cached_at=mock.ANY,
                        )
                    )
                ],
                LivepatchSupport.UNKNOWN,
            ),
        ],
    )
    @mock.patch(M_PATH + "state_files.livepatch_support_cache.write")
    @mock.patch(M_PATH + "UALivepatchClient.is_kernel_supported")
    def test_on_supported_kernel_api(
        self,
        m_is_kernel_supported,
        m_cache_write,
        args,
        is_kernel_supported,
        expected_cache_write_call_args,
        expected,
    ):
        m_is_kernel_supported.return_value = is_kernel_supported
        assert _on_supported_kernel_api(*args) == expected
        assert m_cache_write.call_args_list == expected_cache_write_call_args

    @pytest.mark.parametrize(
        [
            "cli_result",
            "get_kernel_info_result",
            "standardize_arch_name_result",
            "get_release_info_result",
            "cache_result",
            "api_result",
            "cache_call_args",
            "api_call_args",
            "expected",
        ],
        [
            # cli result supported
            (
                LivepatchSupport.SUPPORTED,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                LivepatchSupport.SUPPORTED,
            ),
            # cli result unsupported
            (
                LivepatchSupport.UNSUPPORTED,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                LivepatchSupport.UNSUPPORTED,
            ),
            # cli result upgrade-required
            (
                LivepatchSupport.KERNEL_UPGRADE_REQUIRED,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                LivepatchSupport.KERNEL_UPGRADE_REQUIRED,
            ),
            # cli result eol
            (
                LivepatchSupport.KERNEL_EOL,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                LivepatchSupport.KERNEL_EOL,
            ),
            # cli result definite unknown
            (
                LivepatchSupport.UNKNOWN,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                LivepatchSupport.UNKNOWN,
            ),
            # insufficient kernel info
            (
                None,
                system.KernelInfo(
                    uname_machine_arch="",
                    uname_release="",
                    proc_version_signature_version="",
                    build_date=None,
                    flavor=None,
                    major=5,
                    minor=6,
                    abi=7,
                    patch=None,
                ),
                None,
                None,
                None,
                None,
                [],
                [],
                LivepatchSupport.UNKNOWN,
            ),
            # cache result true
            (
                None,
                system.KernelInfo(
                    uname_machine_arch="",
                    uname_release="",
                    proc_version_signature_version="",
                    build_date=None,
                    flavor="generic",
                    major=5,
                    minor=6,
                    abi=7,
                    patch=None,
                ),
                "amd64",
                mock.MagicMock(series="xenial"),
                (True, True),
                None,
                [mock.call("5.6", "generic", "amd64", "xenial")],
                [],
                LivepatchSupport.SUPPORTED,
            ),
            # cache result false
            (
                None,
                system.KernelInfo(
                    uname_machine_arch="",
                    uname_release="",
                    proc_version_signature_version="",
                    build_date=None,
                    flavor="generic",
                    major=5,
                    minor=6,
                    abi=7,
                    patch=None,
                ),
                "amd64",
                mock.MagicMock(series="xenial"),
                (True, False),
                None,
                [mock.call("5.6", "generic", "amd64", "xenial")],
                [],
                LivepatchSupport.UNSUPPORTED,
            ),
            # cache result none
            (
                None,
                system.KernelInfo(
                    uname_machine_arch="",
                    uname_release="",
                    proc_version_signature_version="",
                    build_date=None,
                    flavor="generic",
                    major=5,
                    minor=6,
                    abi=7,
                    patch=None,
                ),
                "amd64",
                mock.MagicMock(series="xenial"),
                (True, None),
                None,
                [mock.call("5.6", "generic", "amd64", "xenial")],
                [],
                LivepatchSupport.UNKNOWN,
            ),
            # api result true
            (
                None,
                system.KernelInfo(
                    uname_machine_arch="",
                    uname_release="",
                    proc_version_signature_version="",
                    build_date=None,
                    flavor="generic",
                    major=5,
                    minor=6,
                    abi=7,
                    patch=None,
                ),
                "amd64",
                mock.MagicMock(series="xenial"),
                (False, None),
                LivepatchSupport.SUPPORTED,
                [mock.call("5.6", "generic", "amd64", "xenial")],
                [mock.call("5.6", "generic", "amd64", "xenial", None)],
                LivepatchSupport.SUPPORTED,
            ),
        ],
    )
    @mock.patch(M_PATH + "_on_supported_kernel_api")
    @mock.patch(M_PATH + "_on_supported_kernel_cache")
    @mock.patch(M_PATH + "system.get_release_info")
    @mock.patch(M_PATH + "util.standardize_arch_name")
    @mock.patch(M_PATH + "system.get_kernel_info")
    @mock.patch(M_PATH + "_on_supported_kernel_cli")
    def test_on_supported_kernel(
        self,
        m_supported_cli,
        m_get_kernel_info,
        m_standardize_arch_name,
        m_get_release_info,
        m_supported_cache,
        m_supported_api,
        cli_result,
        get_kernel_info_result,
        standardize_arch_name_result,
        get_release_info_result,
        cache_result,
        api_result,
        cache_call_args,
        api_call_args,
        expected,
    ):
        m_supported_cli.return_value = cli_result
        m_get_kernel_info.return_value = get_kernel_info_result
        m_standardize_arch_name.return_value = standardize_arch_name_result
        m_get_release_info.return_value = get_release_info_result
        m_supported_cache.return_value = cache_result
        m_supported_api.return_value = api_result
        assert on_supported_kernel.__wrapped__() == expected
        assert m_supported_cache.call_args_list == cache_call_args
        assert m_supported_api.call_args_list == api_call_args


class TestConfigureLivepatchProxy:
    @pytest.mark.parametrize(
        "http_proxy,https_proxy,retry_sleeps",
        (
            ("http_proxy", "https_proxy", [1, 2]),
            ("http_proxy", "", None),
            ("", "https_proxy", [1, 2]),
            ("http_proxy", None, [1, 2]),
            (None, "https_proxy", None),
            (None, None, [1, 2]),
        ),
    )
    @mock.patch("uaclient.system.subp")
    def test_configure_livepatch_proxy(
        self, m_subp, http_proxy, https_proxy, retry_sleeps, capsys, event
    ):
        configure_livepatch_proxy(http_proxy, https_proxy, retry_sleeps)
        expected_calls = []
        if http_proxy:
            expected_calls.append(
                mock.call(
                    [
                        LIVEPATCH_CMD,
                        "config",
                        "http-proxy={}".format(http_proxy),
                    ],
                    retry_sleeps=retry_sleeps,
                )
            )

        if https_proxy:
            expected_calls.append(
                mock.call(
                    [
                        LIVEPATCH_CMD,
                        "config",
                        "https-proxy={}".format(https_proxy),
                    ],
                    retry_sleeps=retry_sleeps,
                )
            )

        assert m_subp.call_args_list == expected_calls

        out, _ = capsys.readouterr()
        if http_proxy or https_proxy:
            assert out.strip() == messages.SETTING_SERVICE_PROXY.format(
                service=LivepatchEntitlement.title
            )

    @pytest.mark.parametrize(
        "key, subp_return_value, expected_ret",
        [
            ("http-proxy", ("nonsense", ""), None),
            ("http-proxy", ("", "nonsense"), None),
            (
                "http-proxy",
                (
                    """\
http-proxy: ""
https-proxy: ""
no-proxy: ""
remote-server: https://livepatch.canonical.com
ca-certs: ""
check-interval: 60  # minutes""",
                    "",
                ),
                None,
            ),
            (
                "http-proxy",
                (
                    """\
http-proxy: one
https-proxy: two
no-proxy: ""
remote-server: https://livepatch.canonical.com
ca-certs: ""
check-interval: 60  # minutes""",
                    "",
                ),
                "one",
            ),
            (
                "https-proxy",
                (
                    """\
http-proxy: one
https-proxy: two
no-proxy: ""
remote-server: https://livepatch.canonical.com
ca-certs: ""
check-interval: 60  # minutes""",
                    "",
                ),
                "two",
            ),
            (
                "nonexistentkey",
                (
                    """\
http-proxy: one
https-proxy: two
no-proxy: ""
remote-server: https://livepatch.canonical.com
ca-certs: ""
check-interval: 60  # minutes""",
                    "",
                ),
                None,
            ),
        ],
    )
    @mock.patch("uaclient.system.subp")
    def test_get_config_option_value(
        self, m_util_subp, key, subp_return_value, expected_ret
    ):
        m_util_subp.return_value = subp_return_value
        ret = get_config_option_value(key)
        assert ret == expected_ret
        assert [
            mock.call([LIVEPATCH_CMD, "config"])
        ] == m_util_subp.call_args_list


class TestUnconfigureLivepatchProxy:
    @pytest.mark.parametrize(
        "livepatch_installed, protocol_type, retry_sleeps",
        (
            (True, "http", None),
            (True, "https", [1]),
            (True, "http", []),
            (False, "http", None),
        ),
    )
    @mock.patch("uaclient.system.which")
    @mock.patch("uaclient.system.subp")
    def test_unconfigure_livepatch_proxy(
        self, subp, which, livepatch_installed, protocol_type, retry_sleeps
    ):
        if livepatch_installed:
            which.return_value = LIVEPATCH_CMD
        else:
            which.return_value = None
        kwargs = {"protocol_type": protocol_type}
        if retry_sleeps is not None:
            kwargs["retry_sleeps"] = retry_sleeps
        assert None is unconfigure_livepatch_proxy(**kwargs)
        if livepatch_installed:
            expected_calls = [
                mock.call(
                    [LIVEPATCH_CMD, "config", protocol_type + "-proxy="],
                    retry_sleeps=retry_sleeps,
                )
            ]
        else:
            expected_calls = []
        assert expected_calls == subp.call_args_list



#!/usr/bin/env python3

"""
Patch status.json file to avoid tracebacks for non-root from `ua status`

Between 27.1 and 27.2 status.json broke schema key backward-compatibility.

Running `ua status` as non-root users does not generate a new status.json with
the updated status.json schema so the logic which renders ua status
information will traceback on KeyErrors.

Since we haven't introduced official schema versioning in Pro client yet,
this patch is a stop-gap until official schema version handling is
delivered.
"""

import copy
import json
import logging

from uaclient import config, defaults, log, system, util

LOG = logging.getLogger("ubuntupro.lib.patch_status_json")


def patch_status_json_schema_0_1(status_file: str):
    """Patch incompatible status.json file schema to align with version 0.1."""
    content = system.load_file(status_file)
    try:
        status = json.loads(content, cls=util.DatetimeAwareJSONDecoder)
    except ValueError as e:
        LOG.warning(
            "Unable to patch /var/lib/ubuntu-advantage/status.json: %s", str(e)
        )
        return
    new_status = copy.deepcopy(status)
    if float(status.get("_schema_version", 0)) >= 0.1:
        return  # Already have version 0.1 from daily PPA
    LOG.debug("Patching /var/lib/ubuntu-advantage/status.json schema")
    new_status["_schema_version"] = "0.1"
    new_status["account"] = {
        "name": new_status.pop("account", ""),
        "id": new_status.pop("account-id", ""),
    }
    new_status["contract"] = {
        "tech_support_level": new_status.pop("techSupportLevel", "n/a"),
        "name": new_status.pop("subscription", ""),
        "id": new_status.pop("subscription-id", ""),
    }
    if "configStatus" in new_status:
        new_status["execution_status"] = new_status.pop("configStatus")
    if "configStatusDetails" in new_status:
        new_status["execution_details"] = new_status.pop("configStatusDetails")
    try:
        status_content = json.dumps(
            new_status, cls=util.DatetimeAwareJSONEncoder
        )
    except ValueError as e:
        LOG.warning(
            "Unable to patch /var/lib/ubuntu-advantage/status.json: %s", str(e)
        )
    system.write_file(status_file, status_content)


if __name__ == "__main__":
    log.setup_cli_logging(logging.DEBUG, defaults.CONFIG_DEFAULTS["log_level"])
    cfg = config.UAConfig()
    log.setup_cli_logging(cfg.log_level, cfg.log_file)
    patch_status_json_schema_0_1(
        status_file="/var/lib/ubuntu-advantage/status.json"
    )



from typing import List

from uaclient import (
    daemon,
    entitlements,
    exceptions,
    lock,
    messages,
    timer,
    util,
)
from uaclient.api import ProgressWrapper
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo, ErrorWarningObject
from uaclient.api.u.pro.security.status.reboot_required.v1 import (
    _reboot_required,
)
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    StringDataValue,
    data_list,
)
from uaclient.files import machine_token, notices, state_files
from uaclient.timer.update_messaging import update_motd_messages


class DetachResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "disabled",
            data_list(StringDataValue),
            doc="The services disabled during the detach operation",
        ),
        Field(
            "reboot_required",
            BoolDataValue,
            doc=(
                "True if the system requires a reboot after the detach"
                " operation"
            ),
        ),
    ]

    def __init__(self, disabled: List[str], reboot_required: bool):
        self.disabled = disabled
        self.reboot_required = reboot_required


def detach() -> DetachResult:
    return _detach(UAConfig())


def _detach(cfg: UAConfig) -> DetachResult:
    """
    This endpoint allows the user to detach the machine from a Pro
    subscription.
    """
    if not util.we_are_currently_root():
        raise exceptions.NonRootUserError

    try:
        with lock.RetryLock(
            lock_holder="pro.api.u.pro.detach.v1",
        ):
            ret = _detach_in_lock(cfg)
    except Exception as e:
        lock.clear_lock_file_if_present()
        raise e
    return ret


def _detach_in_lock(cfg: UAConfig) -> DetachResult:
    if not _is_attached(cfg).is_attached:
        return DetachResult(
            disabled=[],
            reboot_required=False,
        )

    disabled = []
    warnings = []  # type: List[ErrorWarningObject]
    for ent_name in entitlements.entitlements_disable_order(cfg):
        try:
            ent = entitlements.entitlement_factory(cfg=cfg, name=ent_name)
        except exceptions.EntitlementNotFoundError:
            continue

        # For detach, we should not consider that a service
        # cannot be disabled because of dependent services,
        # since we are going to disable all of them anyway
        can_disable, _ = ent.can_disable(ignore_dependent_services=True)
        if can_disable:
            ret, reason = ent.disable(ProgressWrapper())
            if not ret:
                if reason and reason.message:
                    msg = reason.message.msg
                    code = reason.message.name
                else:
                    msg = messages.DISABLE_FAILED_TMPL.format(title=ent_name)
                    code = ""

                warnings.append(
                    ErrorWarningObject(
                        title=msg,
                        code=code,
                        meta={"service": ent_name},
                    )
                )
            else:
                disabled.append(ent_name)

    notices.remove(notices.Notice.LIMITED_TO_RELEASE)
    state_files.delete_state_files()
    machine_token.get_machine_token_file().delete()
    update_motd_messages(cfg)
    daemon.start()
    timer.stop()

    reboot_required_result = _reboot_required(cfg)

    result = DetachResult(
        disabled=sorted(disabled),
        reboot_required=reboot_required_result.reboot_required == "yes",
    )
    result.warnings = warnings

    return result


endpoint = APIEndpoint(
    version="v1",
    name="Detach",
    fn=_detach,
    options_cls=None,
)

_doc = {
    "introduced_in": "32",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.detach.v1 import detach

result = detach()
""",  # noqa: E501
    "result_class": DetachResult,
    "exceptions": [
        (
            exceptions.NonRootUserError,
            ("Raised if a non-root user executes this endpoint."),
        ),
    ],
    "example_cli": "pro api u.pro.detach.v1",
    "example_json": """
{
    "disabled": ["service1", "service2"],
    "reboot_required": false
}
""",
}



import mock

from uaclient.api.u.pro.security.status.livepatch_cves.v1 import (
    LivepatchCVEsResult,
    _livepatch_cves,
)

M_PATH = "uaclient.api.u.pro.security.status.livepatch_cves.v1."


@mock.patch(M_PATH + "get_livepatch_fixed_cves")
class TestLivepatchCvesV1:
    def test_empty_livepatch_cves(self, m_cves, FakeConfig):
        m_cves.return_value = []
        result = _livepatch_cves(cfg=FakeConfig())
        assert isinstance(result, LivepatchCVEsResult)
        assert result.fixed_cves == []

    def test_livepatch_cves(self, m_cves, FakeConfig):
        m_cves.return_value = [
            {"name": "CVE-123456", "patched": True},
            {"name": "CVE-45678", "patched": False},
        ]
        result = _livepatch_cves(cfg=FakeConfig())
        assert isinstance(result, LivepatchCVEsResult)
        assert len(result.fixed_cves) == 2
        assert result.fixed_cves[0].name == "CVE-123456"
        assert result.fixed_cves[0].patched is True
        assert result.fixed_cves[1].name == "CVE-45678"
        assert result.fixed_cves[1].patched is False



import copy
import enum
import socket
from collections import defaultdict
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

from uaclient import apt, exceptions, livepatch, messages, system, util
from uaclient.http import serviceclient

CVE_OR_USN_REGEX = (
    r"((CVE|cve)-\d{4}-\d{4,7}$|(USN|usn|LSN|lsn)-\d{1,5}-\d{1,2}$)"
)

API_V1_CVES = "cves.json"
API_V1_CVE_TMPL = "cves/{cve}.json"
API_V1_NOTICES = "notices.json"
API_V1_NOTICE_TMPL = "notices/{notice}.json"

STANDARD_UPDATES_POCKET = "standard-updates"
ESM_INFRA_POCKET = "esm-infra"
ESM_APPS_POCKET = "esm-apps"

BinaryPackageFix = NamedTuple(
    "BinaryPackageFix",
    [
        ("source_pkg", str),
        ("binary_pkg", str),
        ("fixed_version", str),
    ],
)

UnfixedPackage = NamedTuple(
    "UnfixedPackage",
    [
        ("pkg", str),
        ("unfixed_reason", str),
    ],
)


class FixStatus(enum.Enum):
    """
    An enum to represent the system status after fix operation
    """

    class _Value:
        def __init__(self, value: int, msg: str):
            self.value = value
            self.msg = msg

    SYSTEM_NON_VULNERABLE = _Value(0, "fixed")
    SYSTEM_NOT_AFFECTED = _Value(0, "not-affected")
    SYSTEM_STILL_VULNERABLE = _Value(1, "still-affected")
    SYSTEM_VULNERABLE_UNTIL_REBOOT = _Value(2, "affected-until-reboot")

    @property
    def exit_code(self):
        return self.value.value

    def __str__(self):
        return self.value.msg


class UASecurityClient(serviceclient.UAServiceClient):

    url_timeout = 20
    cfg_url_base_attr = "security_url"

    def _get_query_params(
        self, query_params: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Update query params with data from feature config.
        """
        extra_security_params = self.cfg.cfg.get("features", {}).get(
            "extra_security_params", {}
        )

        if query_params:
            query_params.update(extra_security_params)
            return query_params

        return extra_security_params

    @util.retry(socket.timeout, retry_sleeps=[1, 3, 5])
    def request_url(
        self, path, data=None, headers=None, method=None, query_params=None
    ):
        query_params = self._get_query_params(query_params)
        return super().request_url(
            path=path,
            data=data,
            headers=headers,
            method=method,
            query_params=query_params,
            log_response_body=False,
        )

    def get_cves(
        self,
        query: Optional[str] = None,
        priority: Optional[str] = None,
        package: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        component: Optional[str] = None,
        version: Optional[str] = None,
        status: Optional[List[str]] = None,
    ) -> List["CVE"]:
        """Query to match multiple-CVEs.

        @return: List of CVE instances based on the the JSON response.
        """
        query_params = {
            "q": query,
            "priority": priority,
            "package": package,
            "limit": limit,
            "offset": offset,
            "component": component,
            "version": version,
            "status": status,
        }
        response = self.request_url(API_V1_CVES, query_params=query_params)
        if response.code != 200:
            raise exceptions.SecurityAPIError(
                url=API_V1_CVES, code=response.code, body=response.body
            )
        return [
            CVE(client=self, response=cve_md) for cve_md in response.json_list
        ]

    def get_cve(self, cve_id: str) -> "CVE":
        """Query to match single-CVE.

        @return: CVE instance for JSON response from the Security API.
        """
        url = API_V1_CVE_TMPL.format(cve=cve_id)
        response = self.request_url(url)
        if response.code != 200:
            raise exceptions.SecurityAPIError(
                url=url, code=response.code, body=response.body
            )
        return CVE(client=self, response=response.json_dict)

    def get_notices(
        self,
        cves: Optional[str] = None,
        release: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order: Optional[str] = None,
    ) -> List["USN"]:
        """Query to match multiple-USNs.

        @return: Sorted list of USN instances based on the the JSON response.
        """
        query_params = {
            "cves": cves,
            "release": release,
            "limit": limit,
            "offset": offset,
            "order": order,
        }
        response = self.request_url(API_V1_NOTICES, query_params=query_params)
        if response.code != 200:
            raise exceptions.SecurityAPIError(
                url=API_V1_NOTICES, code=response.code, body=response.body
            )

        return sorted(
            [
                USN(client=self, response=usn_md)
                for usn_md in response.json_dict.get("notices", [])
                if (cves is None or cves in usn_md.get("cves_ids", []))
                and usn_md.get("id", "").startswith("USN-")
            ],
            key=lambda x: x.id,
        )

    def get_notice(self, notice_id: str) -> "USN":
        """Query to match single-USN.

        @return: USN instance representing the JSON response.
        """
        url = API_V1_NOTICE_TMPL.format(notice=notice_id)
        response = self.request_url(url)
        if response.code != 200:
            raise exceptions.SecurityAPIError(
                url=url, code=response.code, body=response.body
            )
        return USN(client=self, response=response.json_dict)


# Model for Security API responses
class CVEPackageStatus:
    """Class representing specific CVE PackageStatus on an Ubuntu series"""

    def __init__(self, cve_response: Dict[str, Any]):
        self.response = cve_response

    @property
    def description(self):
        return self.response["description"]

    @property
    def fixed_version(self):
        return self.description

    @property
    def pocket(self):
        return self.response["pocket"]

    @property
    def release_codename(self):
        return self.response["release_codename"]

    @property
    def status(self):
        return self.response["status"]

    @property
    def status_message(self):
        if self.status == "needed":
            return messages.SECURITY_CVE_STATUS_NEEDED
        elif self.status == "needs-triage":
            return messages.SECURITY_CVE_STATUS_TRIAGE
        elif self.status == "pending":
            return messages.SECURITY_CVE_STATUS_PENDING
        elif self.status in ("ignored", "deferred"):
            return messages.SECURITY_CVE_STATUS_IGNORED
        elif self.status == "DNE":
            return messages.SECURITY_CVE_STATUS_DNE
        elif self.status == "not-affected":
            return messages.SECURITY_CVE_STATUS_NOT_AFFECTED
        elif self.status == "released":
            return messages.SECURITY_FIX_RELEASE_STREAM.format(
                fix_stream=self.pocket_source
            )
        return messages.SECURITY_CVE_STATUS_UNKNOWN.format(status=self.status)

    @property
    def requires_ua(self) -> bool:
        """Return True if the package requires an active Pro subscription."""
        return bool(
            self.pocket_source
            != messages.SECURITY_UBUNTU_STANDARD_UPDATES_POCKET
        )

    @property
    def pocket_source(self):
        """Human-readable string representing where the fix is published."""
        if self.pocket == "esm-infra":
            fix_source = messages.SECURITY_UA_INFRA_POCKET
        elif self.pocket == "esm-apps":
            fix_source = messages.SECURITY_UA_APPS_POCKET
        elif self.pocket in ("updates", "security"):
            fix_source = messages.SECURITY_UBUNTU_STANDARD_UPDATES_POCKET
        else:
            # TODO(GH: #1376 drop this when esm* pockets supplied by API)
            if "esm" in self.fixed_version:
                fix_source = messages.SECURITY_UA_INFRA_POCKET
            else:
                fix_source = messages.SECURITY_UBUNTU_STANDARD_UPDATES_POCKET
        return fix_source


class CVE:
    """Class representing CVE response from the SecurityClient"""

    def __init__(self, client: UASecurityClient, response: Dict[str, Any]):
        self.response = response
        self.client = client

    def __eq__(self, other) -> bool:
        if not isinstance(other, CVE):
            return False
        return self.response == other.response

    @property
    def id(self):
        return self.response.get("id", "UNKNOWN_CVE_ID").upper()

    @property
    def notices_ids(self) -> List[str]:
        return self.response.get("notices_ids", [])

    @property
    def notices(self) -> List["USN"]:
        """Return a list of USN instances from API response 'notices'.

        Cache the value to avoid extra work on multiple calls.
        """
        if not hasattr(self, "_notices"):
            self._notices = sorted(
                [
                    USN(self.client, notice)
                    for notice in self.response.get("notices", [])
                    if notice and notice.get("id", "").startswith("USN-")
                ],
                key=lambda n: n.id,
                reverse=True,
            )
        return self._notices

    @property
    def description(self):
        return self.response.get("description")

    @property
    def packages_status(self) -> Dict[str, CVEPackageStatus]:
        """Dict of package status dicts for the current Ubuntu series.

        Top-level keys are source packages names and each value is a
        CVEPackageStatus object
        """
        if hasattr(self, "_packages_status"):
            return self._packages_status  # type: ignore
        self._packages_status = {}
        series = system.get_release_info().series
        for package in self.response["packages"]:
            for pkg_status in package["statuses"]:
                if pkg_status["release_codename"] == series:
                    self._packages_status[package["name"]] = CVEPackageStatus(
                        pkg_status
                    )
        return self._packages_status


class USN:
    """Class representing USN response from the SecurityClient"""

    def __init__(self, client: UASecurityClient, response: Dict[str, Any]):
        self.response = response
        self.client = client

    def __eq__(self, other) -> bool:
        if not isinstance(other, USN):
            return False
        return self.response == other.response

    @property
    def id(self) -> str:
        return self.response.get("id", "UNKNOWN_USN_ID").upper()

    @property
    def cves_ids(self) -> List[str]:
        """List of CVE IDs related to this USN."""
        return self.response.get("cves_ids", [])

    @property
    def cves(self) -> List[CVE]:
        """List of CVE instances based on API response 'cves' key.

        Cache the values to avoid extra work for multiple call-sites.
        """
        if not hasattr(self, "_cves"):
            self._cves = sorted(
                [
                    CVE(self.client, cve)
                    for cve in self.response.get("cves", [])
                ],
                key=lambda n: n.id,
                reverse=True,
            )  # type: List[CVE]
        return self._cves

    @property
    def title(self):
        return self.response.get("title")

    @property
    def references(self):
        return self.response.get("references")

    @property
    def release_packages(self) -> Dict[str, Dict[str, Dict[str, str]]]:
        """Binary package information available for this release.


        Reformat the USN.release_packages response to key it based on source
        package name and related binary package names.

        :return: Dict keyed by source package name. The second-level key will
            be binary package names generated from that source package and the
            values will be the dict response from USN.release_packages for
            that binary package. The binary metadata contains the following
            keys: name, version.
            Optional additional keys: pocket and component.
        """
        if hasattr(self, "_release_packages"):
            return self._release_packages
        series = system.get_release_info().series
        self._release_packages = {}  # type: Dict[str, Dict[str, Any]]
        # Organize source and binary packages under a common source package key
        for pkg in self.response.get("release_packages", {}).get(series, []):
            if pkg.get("is_source"):
                # Create a "source" key under src_pkg_name with API response
                if pkg["name"] in self._release_packages:
                    if "source" in self._release_packages[pkg["name"]]:
                        raise exceptions.SecurityAPIMetadataError(
                            error_msg=(
                                "{usn} metadata defines duplicate source"
                                " packages {pkg}"
                            ).format(usn=self.id, pkg=pkg["name"]),
                            issue=self.id,
                            extra_info="",
                        )
                    self._release_packages[pkg["name"]]["source"] = pkg
                else:
                    self._release_packages[pkg["name"]] = {"source": pkg}
            else:
                # is_source == False or None, then this is a binary package.
                # If processed before a source item, the top-level key will
                # not exist yet.
                # TODO(GH: 1465: determine if this is expected on kern pkgs)
                if not pkg.get("source_link"):
                    raise exceptions.SecurityAPIMetadataError(
                        error_msg=(
                            "{issue} metadata does not define release_packages"
                            " source_link for {bin_pkg}."
                        ).format(issue=self.id, bin_pkg=pkg["name"]),
                        issue=self.id,
                        extra_info="",
                    )
                elif "/" not in pkg["source_link"]:
                    raise exceptions.SecurityAPIMetadataError(
                        error_msg=(
                            "{issue} metadata has unexpected release_packages"
                            " source_link value for {bin_pkg}: {link}"
                        ).format(
                            issue=self.id,
                            bin_pkg=pkg["name"],
                            link=pkg["source_link"],
                        ),
                        issue=self.id,
                        extra_info="",
                    )
                source_pkg_name = pkg["source_link"].split("/")[-1]
                if source_pkg_name not in self._release_packages:
                    self._release_packages[source_pkg_name] = {}
                self._release_packages[source_pkg_name][pkg["name"]] = pkg
        return self._release_packages


def get_cve_affected_source_packages_status(
    cve: CVE, installed_packages: Dict[str, Dict[str, str]]
) -> Dict[str, CVEPackageStatus]:
    """Get a dict of any CVEPackageStatuses affecting this Ubuntu release.

    :return: Dict of active CVEPackageStatus keyed by source package names.
    """
    affected_pkg_versions = {}
    for source_pkg, package_status in cve.packages_status.items():
        if package_status.status == "not-affected":
            continue
        if source_pkg in installed_packages:
            affected_pkg_versions[source_pkg] = package_status
    return affected_pkg_versions


def query_installed_source_pkg_versions() -> Dict[str, Dict[str, str]]:
    """Return a dict of all source packages installed on the system.

    The dict keys will be source package name: "krb5". The value will be a dict
    with keys binary_pkg and version.
    """
    status_field = "${db:Status-Status}"
    out, _err = system.subp(
        [
            "dpkg-query",
            "-f=${Package},${Source},${Version}," + status_field + "\n",
            "-W",
        ]
    )
    installed_packages = {}  # type: Dict[str, Dict[str, str]]
    for pkg_line in out.splitlines():
        pkg_name, source_pkg_name, pkg_version, status = pkg_line.split(",")
        if not source_pkg_name:
            # some package don't define the Source
            source_pkg_name = pkg_name
        if "installed" not in status:
            continue
        if source_pkg_name in installed_packages:
            installed_packages[source_pkg_name][pkg_name] = pkg_version
        else:
            installed_packages[source_pkg_name] = {pkg_name: pkg_version}
    return installed_packages


def get_related_usns(usn, client):
    """For a give usn, get the related USNs for it.

    For each CVE associated with the given USN, we capture
    other USNs that are related to the CVE. We consider those
    USNs related to the original USN.
    """

    # If the usn does not have any associated cves on it,
    # we cannot establish a relation between USNs
    if not usn.cves:
        return []

    related_usns = {}
    for cve in usn.cves:
        for related_usn_id in cve.notices_ids:
            # We should ignore any other item that is not a USN
            # For example, LSNs
            if not related_usn_id.startswith("USN-"):
                continue
            if related_usn_id == usn.id:
                continue
            if related_usn_id not in related_usns:
                related_usns[related_usn_id] = client.get_notice(
                    notice_id=related_usn_id
                )

    return list(sorted(related_usns.values(), key=lambda x: x.id))


def get_affected_packages_from_cves(cves, installed_packages):
    affected_pkgs = {}  # type: Dict[str, CVEPackageStatus]

    for cve in cves:
        for pkg_name, pkg_status in get_cve_affected_source_packages_status(
            cve, installed_packages
        ).items():
            if pkg_name not in affected_pkgs:
                affected_pkgs[pkg_name] = pkg_status
            else:
                current_ver = affected_pkgs[pkg_name].fixed_version
                if (
                    apt.version_compare(current_ver, pkg_status.fixed_version)
                    > 0
                ):
                    affected_pkgs[pkg_name] = pkg_status

    return affected_pkgs


def get_affected_packages_from_usn(usn, installed_packages):
    affected_pkgs = {}  # type: Dict[str, CVEPackageStatus]
    for pkg_name, pkg_info in usn.release_packages.items():
        if pkg_name not in installed_packages:
            continue

        cve_response = defaultdict(str)
        cve_response["status"] = "released"
        # Here we are assuming that the pocket will be the same one across
        # all the different binary packages.
        all_pockets = {
            pkg_bin_info["pocket"]
            for _, pkg_bin_info in pkg_info.items()
            if pkg_bin_info.get("pocket")
        }
        if not all_pockets:
            raise exceptions.SecurityAPIMetadataError(
                error_msg=(
                    "{} metadata defines no pocket information for "
                    "any release packages."
                ).format(usn.id),
                issue=usn.id,
                extra_info="",
            )
        cve_response["pocket"] = all_pockets.pop()

        affected_pkgs[pkg_name] = CVEPackageStatus(cve_response=cve_response)

    return affected_pkgs


def get_usn_affected_packages_status(
    usn: USN, installed_packages: Dict[str, Dict[str, str]]
) -> Dict[str, CVEPackageStatus]:
    """Walk CVEs related to a USN and return a dict of all affected packages.

    :return: Dict keyed on source package name, with active CVEPackageStatus
        for the current Ubuntu release.
    """
    if usn.cves:
        return get_affected_packages_from_cves(usn.cves, installed_packages)
    else:
        return get_affected_packages_from_usn(usn, installed_packages)


def override_usn_release_package_status(
    pkg_status: CVEPackageStatus,
    usn_src_released_pkgs: Dict[str, Dict[str, str]],
) -> CVEPackageStatus:
    """Parse release status based on both pkg_status and USN.release_packages.

    Since some source packages in universe are not represented in
    CVEPackageStatus, rely on presence of such source packages in
    usn_src_released_pkgs to represent package as a "released" status.

    :param pkg_status: the CVEPackageStatus for this source package.
    :param usn_src_released_pkgs: The USN.release_packages representing only
       this source package. Normally, release_packages would have data on
       multiple source packages.

    :return: Tuple of:
        human-readable status message, boolean whether released,
        boolean whether the fix requires access to UA
    """

    usn_pkg_status = copy.deepcopy(pkg_status)
    if usn_src_released_pkgs and usn_src_released_pkgs.get("source"):
        usn_pkg_status.response["status"] = "released"
        usn_pkg_status.response["description"] = usn_src_released_pkgs[
            "source"
        ]["version"]
        for pkg_name, usn_released_pkg in usn_src_released_pkgs.items():
            # Copy the pocket from any valid binary package
            pocket = usn_released_pkg.get("pocket")
            if pocket:
                usn_pkg_status.response["pocket"] = pocket
                break
    return usn_pkg_status


def group_by_usn_package_status(affected_pkg_status, usn_released_pkgs):
    status_groups = {}  # type: Dict[str, List[Tuple[str, CVEPackageStatus]]]
    for src_pkg, pkg_status in sorted(affected_pkg_status.items()):
        usn_released_src = usn_released_pkgs.get(src_pkg, {})
        usn_pkg_status = override_usn_release_package_status(
            pkg_status, usn_released_src
        )
        status_group = usn_pkg_status.status.replace("ignored", "deferred")
        if status_group not in status_groups:
            status_groups[status_group] = []
        status_groups[status_group].append((src_pkg, usn_pkg_status))
    return status_groups


def merge_usn_released_binary_package_versions(
    usns: List[USN], beta_pockets: Dict[str, bool]
) -> Dict[str, Dict[str, Dict[str, str]]]:
    """Walk related USNs, merging the released binary package versions.

    For each USN, iterate over release_packages to collect released binary
        package names and required fix version. If multiple related USNs
        require different version fixes to the same binary package, track the
        maximum version required across all USNs.

    :param usns: List of USN response instances from which to calculate merge.
    :param beta_pockets: Dict keyed on service name: esm-infra, esm-apps
        the values of which will be true of USN response instances
        from which to calculate merge.

    :return: Dict keyed by source package name. Under each source package will
        be a dict with binary package name as keys and binary package metadata
        as the value.
    """
    usn_pkg_versions = {}
    for usn in usns:
        # Aggregate USN.release_package binary versions into usn_pkg_versions
        for src_pkg, binary_pkg_versions in usn.release_packages.items():
            public_bin_pkg_versions = {
                bin_pkg_name: bin_pkg_md
                for bin_pkg_name, bin_pkg_md in binary_pkg_versions.items()
                if False
                is beta_pockets.get(bin_pkg_md.get("pocket", "None"), False)
            }
            if src_pkg not in usn_pkg_versions and public_bin_pkg_versions:
                usn_pkg_versions[src_pkg] = public_bin_pkg_versions
            elif src_pkg in usn_pkg_versions:
                # Since src_pkg exists, only record this USN's binary version
                # when it is greater than the previous version in usn_src_pkg.
                usn_src_pkg = usn_pkg_versions[src_pkg]
                for bin_pkg, binary_pkg_md in public_bin_pkg_versions.items():
                    if bin_pkg not in usn_src_pkg:
                        usn_src_pkg[bin_pkg] = binary_pkg_md
                    else:
                        prev_version = usn_src_pkg[bin_pkg]["version"]
                        current_version = binary_pkg_md["version"]
                        if (
                            apt.version_compare(current_version, prev_version)
                            > 0
                        ):
                            # binary_version is greater than prev_version
                            usn_src_pkg[bin_pkg] = binary_pkg_md
    return usn_pkg_versions


def _check_cve_fixed_by_livepatch(
    issue_id: str,
) -> Tuple[Optional[FixStatus], Optional[str]]:
    # Check livepatch status for CVE in fixes before checking CVE api
    lp_status = livepatch.status()
    if (
        lp_status is not None
        and lp_status.livepatch is not None
        and lp_status.livepatch.fixes is not None
    ):
        for fix in lp_status.livepatch.fixes:
            if fix.name == issue_id.lower() and fix.patched:
                version = lp_status.livepatch.version or "N/A"
                return (FixStatus.SYSTEM_NON_VULNERABLE, version)

    return (None, None)


def status_message(status, pocket_source: Optional[str] = None):
    if status == "needed":
        return messages.SECURITY_CVE_STATUS_NEEDED
    elif status == "needs-triage":
        return messages.SECURITY_CVE_STATUS_TRIAGE
    elif status == "pending":
        return messages.SECURITY_CVE_STATUS_PENDING
    elif status in ("ignored", "deferred"):
        return messages.SECURITY_CVE_STATUS_IGNORED
    elif status == "DNE":
        return messages.SECURITY_CVE_STATUS_DNE
    elif status == "not-affected":
        return messages.SECURITY_CVE_STATUS_NOT_AFFECTED
    elif status == "released" and pocket_source:
        return messages.SECURITY_FIX_RELEASE_STREAM.format(
            fix_stream=pocket_source
        )
    return messages.SECURITY_CVE_STATUS_UNKNOWN.format(status=status)


def get_expected_overall_status(
    current_fix_status: str, fix_status: str
) -> str:
    if not current_fix_status:
        return fix_status

    if fix_status in (
        FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
        FixStatus.SYSTEM_NOT_AFFECTED.value.msg,
    ):
        if (
            current_fix_status == FixStatus.SYSTEM_NOT_AFFECTED.value.msg
            and current_fix_status != fix_status
        ):
            return fix_status
        else:
            return current_fix_status
    else:
        # This means the system is still affected and we must
        # priotize this as the global state
        return fix_status



from typing import List, Optional, Tuple

from uaclient import messages


class InPlaceUpgradeNotSupportedError(Exception):
    pass


class IsProLicensePresentError(Exception):
    pass


class CancelProLicensePolling(IsProLicensePresentError):
    pass


class DelayProLicensePolling(IsProLicensePresentError):
    pass


class ProcessExecutionError(IOError):
    def __init__(
        self,
        cmd: str,
        exit_code: Optional[int] = None,
        stdout: str = "",
        stderr: str = "",
    ) -> None:
        self.stdout = stdout
        self.stderr = stderr
        self.exit_code = exit_code
        if not exit_code:
            message = messages.SUBP_INVALID_COMMAND.format(cmd=cmd)
        else:
            message = messages.SUBP_COMMAND_FAILED.format(
                cmd=cmd, exit_code=exit_code, stderr=stderr
            )
        super().__init__(message)


class UbuntuProError(Exception):
    """
    Base class for all of our custom errors.
    All possible exceptions from our API should extend this class.
    """

    _msg = None  # type: messages.NamedMessage
    _formatted_msg = None  # type: messages.FormattedNamedMessage

    exit_code = 1

    def __init__(self, **kwargs) -> None:
        if self._formatted_msg is not None:
            self.named_msg = self._formatted_msg.format(
                **kwargs
            )  # type: messages.NamedMessage
        else:
            self.named_msg = self._msg

        self.additional_info = kwargs

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def msg(self):
        return self.named_msg.msg

    @property
    def msg_code(self):
        return self.named_msg.name

    def __str__(self):
        return self.named_msg.msg


# UserFacingError is the old name of our base exception class.
# Other systems may import and rely on this name.
# Leave this here for backwards compatibility.
UserFacingError = UbuntuProError


class AnonymousUbuntuProError(UbuntuProError):
    """
    For cases where we have a NamedMessage of an Error already but it isn't in
    exception form, and we need to raise it as an exception. This generally
    should only be used in cli-only code.
    """

    def __init__(self, named_msg: messages.NamedMessage, **kwargs):
        super().__init__(**kwargs)
        self.named_msg = named_msg


###############################################################################
#                              APT                                            #
###############################################################################


class APTProcessConflictError(UbuntuProError):
    _msg = messages.E_APT_PROCESS_CONFLICT


class APTInvalidRepoError(UbuntuProError):
    _formatted_msg = messages.E_APT_UPDATE_INVALID_URL_CONFIG


class APTUpdateProcessConflictError(UbuntuProError):
    _msg = messages.E_APT_UPDATE_PROCESS_CONFLICT


class APTUpdateInvalidRepoError(UbuntuProError):
    _formatted_msg = messages.E_APT_UPDATE_INVALID_REPO


class APTUpdateFailed(UbuntuProError):
    _formatted_msg = messages.E_APT_UPDATE_FAILED


class APTInstallProcessConflictError(UbuntuProError):
    _msg = messages.E_APT_INSTALL_PROCESS_CONFLICT


class APTInstallInvalidRepoError(UbuntuProError):
    _formatted_msg = messages.E_APT_INSTALL_INVALID_REPO


class APTInvalidCredentials(UbuntuProError):
    _formatted_msg = messages.E_APT_INVALID_CREDENTIALS


class APTTimeout(UbuntuProError):
    _formatted_msg = messages.E_APT_TIMEOUT


class APTUnexpectedError(UbuntuProError):
    _formatted_msg = messages.E_APT_UNEXPECTED_ERROR


class APTCommandTimeout(UbuntuProError):
    _formatted_msg = messages.E_APT_COMMAND_TIMEOUT


###############################################################################
#                              SNAP                                           #
###############################################################################


class SnapNotInstalledError(UbuntuProError):
    _formatted_msg = messages.E_SNAP_NOT_INSTALLED_ERROR


class UnexpectedSnapdAPIError(UbuntuProError):
    _formatted_msg = messages.E_UNEXPECTED_SNAPD_API_ERROR


class SnapdAPIConnectionRefused(UbuntuProError):
    _msg = messages.E_SNAPD_CONNECTION_REFUSED


class CannotInstallSnapdError(UbuntuProError):
    _msg = messages.E_CANNOT_INSTALL_SNAPD


class ErrorInstallingLivepatch(UbuntuProError):
    _formatted_msg = messages.E_ERROR_INSTALLING_LIVEPATCH


###############################################################################
#                              PROXY/NETWORK                                  #
###############################################################################


class InvalidUrl(UbuntuProError):
    _formatted_msg = messages.E_INVALID_URL


class ProxyNotWorkingError(UbuntuProError):
    _formatted_msg = messages.E_NOT_SETTING_PROXY_NOT_WORKING


class ProxyInvalidUrl(UbuntuProError):
    _formatted_msg = messages.E_NOT_SETTING_PROXY_INVALID_URL


class PycurlRequiredError(UbuntuProError):
    _msg = messages.E_PYCURL_REQUIRED


class PycurlError(UbuntuProError):
    _formatted_msg = messages.E_PYCURL_ERROR


class ProxyAuthenticationFailed(UbuntuProError):
    _msg = messages.E_PROXY_AUTH_FAIL


class ExternalAPIError(UbuntuProError):
    _formatted_msg = messages.E_EXTERNAL_API_ERROR
    code = None  # type: int
    url = None  # type: str
    body = None  # type: str

    def __str__(self):
        return "{}: [{}], {}".format(self.code, self.url, self.body)


class ContractAPIError(ExternalAPIError):
    pass


class SecurityAPIError(ExternalAPIError):
    pass


class PycurlCACertificatesError(UbuntuProError):
    _msg = messages.E_PYCURL_CA_CERTIFICATES

    def __init__(self, url, **kwargs) -> None:
        super().__init__(**kwargs)
        self.url = url


class ConnectivityError(UbuntuProError, IOError):
    _formatted_msg = messages.E_CONNECTIVITY_ERROR

    def __init__(
        self,
        cause: Exception,
        url: str,
    ):
        if getattr(cause, "reason", None):
            cause_error = str(getattr(cause, "reason"))
        else:
            cause_error = str(cause)
        IOError.__init__(self, cause_error)
        UbuntuProError.__init__(self, cause_error=cause_error, url=url)

        # Even though we already set those variable through UbuntuProError
        # we need to set them again to avoid mypy warnings
        self.cause_error = cause_error
        self.url = url


# We are doing that just to keep backwards compatibility
# for our custom UrlError exception
UrlError = ConnectivityError

###############################################################################
#                              ATTACH/ENABLE                                  #
###############################################################################


class ContractExpiredError(UbuntuProError):
    _msg = messages.E_CONTRACT_EXPIRED


class InvalidServiceOpError(UbuntuProError):
    _formatted_msg = messages.E_INVALID_SERVICE_OP_FAILURE


class AlreadyAttachedError(UbuntuProError):
    """An exception to be raised when a command needs an unattached system."""

    exit_code = 2
    _formatted_msg = messages.E_ALREADY_ATTACHED


class AttachError(UbuntuProError):
    """An exception to be raised when we detect a generic attach error."""

    exit_code = 1
    _msg = messages.E_ATTACH_FAILURE


class AttachInvalidConfigFileError(UbuntuProError):
    _formatted_msg = messages.E_ATTACH_CONFIG_READ_ERROR


class AttachInvalidTokenError(UbuntuProError):
    _msg = messages.E_ATTACH_INVALID_TOKEN


class AttachForbiddenExpired(UbuntuProError):
    _formatted_msg = messages.E_ATTACH_FORBIDDEN_EXPIRED


class AttachForbiddenNotYet(UbuntuProError):
    _formatted_msg = messages.E_ATTACH_FORBIDDEN_NOT_YET


class AttachForbiddenNever(UbuntuProError):
    _formatted_msg = messages.E_ATTACH_FORBIDDEN_NEVER


class AttachExpiredToken(UbuntuProError):
    _msg = messages.E_ATTACH_EXPIRED_TOKEN


class MagicAttachTokenAlreadyActivated(UbuntuProError):
    _msg = messages.E_MAGIC_ATTACH_TOKEN_ALREADY_ACTIVATED


class MagicAttachTokenError(UbuntuProError):
    _msg = messages.E_MAGIC_ATTACH_TOKEN_ERROR


class MagicAttachUnavailable(UbuntuProError):
    _msg = messages.E_MAGIC_ATTACH_UNAVAILABLE


class MagicAttachInvalidParam(UbuntuProError):
    _formatted_msg = messages.E_MAGIC_ATTACH_INVALID_PARAM


class MissingAptURLDirective(UbuntuProError):
    """An exception for when the contract server doesn't include aptURL"""

    _formatted_msg = messages.E_MISSING_APT_URL_DIRECTIVE


class UnattachedError(UbuntuProError):
    """An exception to be raised when a machine needs to be attached."""

    _msg = messages.E_UNATTACHED


class UnattachedValidServicesError(UbuntuProError):
    _formatted_msg = messages.E_VALID_SERVICE_FAILURE_UNATTACHED


class UnattachedMixedServicesError(UbuntuProError):
    _formatted_msg = messages.E_MIXED_SERVICES_FAILURE_UNATTACHED


class UnattachedInvalidServicesError(UbuntuProError):
    _formatted_msg = messages.E_INVALID_SERVICE_OP_FAILURE


class EntitlementNotFoundError(UbuntuProError):
    _formatted_msg = messages.E_ENTITLEMENT_NOT_FOUND


class EntitlementsNotEnabledError(UbuntuProError):
    exit_code = 4
    _msg = messages.E_ENTITLEMENTS_NOT_ENABLED_ERROR

    def __init__(
        self, failed_services: List[Tuple[str, messages.NamedMessage]]
    ):
        info_dicts = [
            {"name": f[0], "code": f[1].name, "title": f[1].msg}
            for f in failed_services
        ]
        super().__init__(
            additional_info={"services": info_dicts},
        )


class EntitlementNotEnabledError(UbuntuProError):
    _formatted_msg = messages.E_ENTITLEMENT_NOT_ENABLED_ERROR

    def __init__(self, service: str, reason: messages.NamedMessage):
        super().__init__(
            service=service,
            additional_info={
                "reason": {
                    "code": reason.name,
                    "title": reason.msg,
                    "additional_info": reason.additional_info,
                }
            },
        )


class EntitlementNotDisabledError(UbuntuProError):
    _formatted_msg = messages.E_ENTITLEMENT_NOT_DISABLED_ERROR

    def __init__(self, service: str, reason: messages.NamedMessage):
        super().__init__(
            service=service,
            additional_info={
                "reason": {
                    "code": reason.name,
                    "title": reason.msg,
                    "additional_info": reason.additional_info,
                }
            },
        )


class AttachFailureDefaultServices(EntitlementsNotEnabledError):
    _msg = messages.E_ATTACH_FAILURE_DEFAULT_SERVICES


class AttachFailureUnknownError(EntitlementsNotEnabledError):
    _msg = messages.E_ATTACH_FAILURE_UNEXPECTED


class AttachFailureRestrictedRelease(UbuntuProError):
    _formatted_msg = messages.ATTACH_FAILURE_RESTRICTED_RELEASE


class RepoNoAptKey(UbuntuProError):
    _formatted_msg = messages.E_REPO_NO_APT_KEY


class RepoNoSuites(UbuntuProError):
    _formatted_msg = messages.E_REPO_NO_SUITES


class RepoPinFailNoOrigin(UbuntuProError):
    _formatted_msg = messages.E_REPO_PIN_FAIL_NO_ORIGIN


class InvalidContractDeltasServiceType(UbuntuProError):
    _formatted_msg = messages.E_INVALID_CONTRACT_DELTAS_SERVICE_TYPE


class EntitlementsAPTDirectivesAreNotUnique(UbuntuProError):
    _formatted_msg = messages.E_ENTITLEMENTS_APT_DIRECTIVES_ARE_NOT_UNIQUE


class RequiredServiceStopsEnable(UbuntuProError):
    _formatted_msg = messages.E_REQUIRED_SERVICE_STOPS_ENABLE


class IncompatibleServiceStopsEnable(UbuntuProError):
    _formatted_msg = messages.E_INCOMPATIBLE_SERVICE_STOPS_ENABLE


class DependentServiceStopsDisable(UbuntuProError):
    _formatted_msg = messages.E_DEPENDENT_SERVICE_STOPS_DISABLE


class LandscapeConfigFailed(UbuntuProError):
    _msg = messages.E_LANDSCAPE_CONFIG_FAILED


class NonInteractiveKernelPurgeDisallowed(UbuntuProError):
    _msg = messages.E_NON_INTERACTIVE_KERNEL_PURGE_DISALLOWED


###############################################################################
#                              CLOUD AUTO-ATTACH                              #
###############################################################################


class InvalidProImage(UbuntuProError):
    _formatted_msg = messages.E_INVALID_PRO_IMAGE
    error_msg = None  # type: str


class CloudMetadataError(UbuntuProError):
    _formatted_msg = messages.E_CLOUD_METADATA_ERROR


class GCPServiceAccountError(CloudMetadataError):
    """An exception raised when GCP service account is not enabled"""

    _formatted_msg = messages.E_GCP_SERVICE_ACCT_NOT_ENABLED_ERROR


class AWSNoValidIMDS(UbuntuProError):
    _formatted_msg = messages.E_AWS_NO_VALID_IMDS


class CloudFactoryError(UbuntuProError):
    pass


class CloudFactoryNoCloudError(CloudFactoryError):
    _msg = messages.E_UNABLE_TO_DETERMINE_CLOUD_TYPE


class CloudFactoryNonViableCloudError(CloudFactoryError):
    _msg = messages.E_UNSUPPORTED_AUTO_ATTACH


class NonAutoAttachImageError(CloudFactoryError):
    """Raised when machine isn't running an auto-attach enabled image"""

    exit_code = 0
    _formatted_msg = messages.E_UNSUPPORTED_AUTO_ATTACH_CLOUD_TYPE


class LXDAutoAttachNotAvailable(UbuntuProError):
    _msg = messages.E_LXD_AUTO_ATTACH_NOT_AVAILABLE


class LXDAutoAttachNotAllowed(UbuntuProError):
    _msg = messages.E_LXD_AUTO_ATTACH_NOT_ALLOWED


###############################################################################
#                              FILES/FORMATS                                  #
###############################################################################


class InvalidFileFormatError(UbuntuProError):
    _formatted_msg = messages.E_INVALID_FILE_FORMAT


class InvalidFileEncodingError(UbuntuProError):
    _formatted_msg = messages.E_INVALID_FILE_ENCODING


class ParsingErrorOnOSReleaseFile(UbuntuProError):
    _formatted_msg = messages.E_ERROR_PARSING_VERSION_OS_RELEASE


class MissingSeriesOnOSReleaseFile(UbuntuProError):
    _formatted_msg = messages.E_MISSING_SERIES_ON_OS_RELEASE


class InvalidLockFile(UbuntuProError):
    _formatted_msg = messages.E_INVALID_LOCK_FILE


class InvalidJson(UbuntuProError):
    _formatted_msg = messages.E_JSON_PARSER_ERROR


class InvalidBooleanConfigValue(UbuntuProError):
    _formatted_msg = messages.E_INVALID_BOOLEAN_CONFIG_VALUE


class InvalidPosIntConfigValue(UbuntuProError):
    _formatted_msg = messages.E_CLI_CONFIG_VALUE_MUST_BE_POS_INT


class InvalidURLConfigValue(UbuntuProError):
    _formatted_msg = messages.E_CONFIG_INVALID_URL


class InvalidFeatureYamlConfigValue(UbuntuProError):
    _formatted_msg = messages.E_CONFIG_NO_YAML_FILE


class InvalidProxyCombinationConfig(UbuntuProError):
    _msg = messages.E_INVALID_PROXY_COMBINATION


class MissingDistroInfoFile(UbuntuProError):
    _msg = messages.E_MISSING_DISTRO_INFO_FILE


class MissingSeriesInDistroInfoFile(UbuntuProError):
    _formatted_msg = messages.E_MISSING_SERIES_IN_DISTRO_INFO_FILE


class UnsupportedManifestFile(UbuntuProError):
    _formatted_msg = messages.E_UNSUPPORTED_MANIFEST_FILE


###############################################################################
#                              CLI                                            #
###############################################################################


class InvalidOptionCombination(UbuntuProError):
    _formatted_msg = messages.E_INVALID_OPTION_COMBINATION


class DepedentOptionError(UbuntuProError):
    _formatted_msg = messages.E_DEPENDENT_OPTION


class NoHelpContent(UbuntuProError):
    _formatted_msg = messages.E_CLI_NO_HELP


class InvalidSecurityIssueIdFormat(UbuntuProError):
    _formatted_msg = messages.E_SECURITY_FIX_CLI_ISSUE_REGEX_FAIL


class InvalidArgChoice(UbuntuProError):
    _formatted_msg = messages.E_CLI_VALID_CHOICES


class EmptyConfigValue(UbuntuProError):
    _formatted_msg = messages.E_CLI_EMPTY_CONFIG_VALUE


class GenericInvalidFormat(UbuntuProError):
    _formatted_msg = messages.E_CLI_EXPECTED_FORMAT


class RefreshConfigFailure(UbuntuProError):
    _msg = messages.E_REFRESH_CONFIG_FAILURE


class RefreshContractFailure(UbuntuProError):
    _msg = messages.E_REFRESH_CONTRACT_FAILURE


class RefreshMessagesFailure(UbuntuProError):
    _msg = messages.E_REFRESH_MESSAGES_FAILURE


class CLIJSONFormatRequireAssumeYes(UbuntuProError):
    _msg = messages.E_JSON_FORMAT_REQUIRE_ASSUME_YES


class CLIAttachTokenArgXORConfig(UbuntuProError):
    _msg = messages.E_ATTACH_TOKEN_ARG_XOR_CONFIG


class CLIAPIOptionsXORData(UbuntuProError):
    _msg = messages.E_API_ERROR_ARGS_AND_DATA_TOGETHER


class PromptDeniedError(UbuntuProError):
    _msg = messages.E_PROMPT_DENIED


###############################################################################
#                              VULNERABILITIES                                #
###############################################################################


class VulnerabilityDataNotFound(UbuntuProError):
    _msg = messages.E_VULNERABILITY_DATA_NOT_FOUND


###############################################################################
#                              MISCELLANEOUS                                  #
###############################################################################


class LockHeldError(UbuntuProError):
    """An exception for when another pro operation is in progress

    :param lock_request: String of the command requesting the lock
    :param lock_holder: String of the command that currently holds the lock
    :param pid: Integer of the process id of the lock_holder
    """

    _formatted_msg = messages.E_LOCK_HELD_ERROR
    pid = None  # type: int


class NonRootUserError(UbuntuProError):
    """An exception to be raised when a user needs to be root."""

    _msg = messages.E_NONROOT_USER


class SecurityAPIMetadataError(UbuntuProError):
    """An exception raised with Security API metadata returns invalid data."""

    _formatted_msg = messages.E_SECURITY_API_INVALID_METADATA


class SecurityIssueNotFound(UbuntuProError):
    _formatted_msg = messages.E_SECURITY_FIX_NOT_FOUND_ISSUE


class GPGKeyNotFound(UbuntuProError):
    _formatted_msg = messages.E_GPG_KEY_NOT_FOUND


class UnknownProcessorType(UbuntuProError):
    _formatted_msg = messages.E_UNKNOWN_PROCESSOR_TYPE


class FeatureNotSupportedOldTokenError(UbuntuProError):
    _formatted_msg = messages.E_FEATURE_NOT_SUPPORTED_OLD_TOKEN


class ETagUnchanged(UbuntuProError):
    _formatted_msg_ = messages.E_ETAG_UNCHANGED


class NoServicesToEnable(UbuntuProError):
    _msg = messages.E_NO_SERVICES_TO_ENABLE



import json
import logging

import requests
import yaml
from behave import then, when
from hamcrest import assert_that, equal_to, not_

from features.steps.files import (
    change_config_key_to_use_value,
    when_i_create_file_with_content,
)
from features.steps.shell import when_i_run_command
from features.util import SUT, process_template_vars
from uaclient import util
from uaclient.defaults import (
    DEFAULT_CONFIG_FILE,
    DEFAULT_PRIVATE_MACHINE_TOKEN_PATH,
)


@when("I update contract to use `{contract_field}` as `{new_value}`")
def when_i_update_contract_field_to_new_value(
    context, contract_field, new_value
):
    new_value = process_template_vars(context, new_value)
    when_i_run_command(
        context,
        'sed -i \'s/"{}": "[^"]*"/"{}": "{}"/g\' {}'.format(
            contract_field,
            contract_field,
            new_value,
            DEFAULT_PRIVATE_MACHINE_TOKEN_PATH,
        ),
        user_spec="with sudo",
    )


@when("I change contract to staging {user_spec}")
def change_contract_endpoint_to_staging(context, user_spec):
    when_i_run_command(
        context,
        "sed -i 's/contracts.can/contracts.staging.can/' {}".format(
            DEFAULT_CONFIG_FILE
        ),
        user_spec,
    )


def change_contract_endpoint_to_production(context, user_spec):
    when_i_run_command(
        context,
        "sed -i 's/contracts.staging.can/contracts.can/' {}".format(
            DEFAULT_CONFIG_FILE
        ),
        user_spec,
    )


@when("I save the `{key}` value from the contract")
def i_save_the_key_value_from_contract(context, key):
    when_i_run_command(
        context,
        "jq -r '.{}' {}".format(key, DEFAULT_PRIVATE_MACHINE_TOKEN_PATH),
        "with sudo",
    )
    output = context.process.stdout.strip()

    if output:
        if not hasattr(context, "saved_values"):
            setattr(context, "saved_values", {})

        context.saved_values[key] = output


def _get_saved_attr(context, key):
    saved_value = getattr(context, "saved_values", {}).get(key)

    if saved_value is None:
        raise AssertionError(
            "Value for key {} was not previously saved\n".format(key)
        )

    return saved_value


@then("I verify that `{key}` value has been updated on the contract")
@then(
    "I verify that `{key}` value has been updated on the contract on the `{machine_name}` machine"  # noqa: E501
)
def i_verify_that_key_value_has_been_updated(context, key, machine_name=SUT):
    saved_value = _get_saved_attr(context, key)
    when_i_run_command(
        context,
        "jq -r '.{}' {}".format(key, DEFAULT_PRIVATE_MACHINE_TOKEN_PATH),
        "with sudo",
        machine_name=machine_name,
    )
    assert_that(context.process.stdout.strip(), not_(equal_to(saved_value)))


@then("I verify that `{key}` value has not been updated on the contract")
def i_verify_that_key_value_has_not_been_updated(context, key):
    saved_value = _get_saved_attr(context, key)
    when_i_run_command(
        context,
        "jq -r '.{}' {}".format(key, DEFAULT_PRIVATE_MACHINE_TOKEN_PATH),
        "with sudo",
    )
    assert_that(context.process.stdout.strip(), equal_to(saved_value))


@when("I restore the saved `{key}` value on contract")
def i_restore_the_saved_key_value_on_contract(context, key):
    saved_value = _get_saved_attr(context, key)
    when_i_update_contract_field_to_new_value(
        context=context,
        contract_field=key.split(".")[-1],
        new_value=saved_value,
    )


@when("I set the machine token overlay to the following yaml")
def when_i_set_the_machine_token_overlay(context):
    json_text = json.dumps(
        yaml.safe_load(context.text), cls=util.DatetimeAwareJSONEncoder
    )
    when_i_create_file_with_content(
        context,
        "/var/lib/ubuntu-advantage/machine-token-overlay.json",
        text=json_text,
    )
    change_config_key_to_use_value(
        context,
        "features",
        "{ machine_token_overlay: "
        "/var/lib/ubuntu-advantage/machine-token-overlay.json}",
    )


@when("I set the test contract expiration date to `{date_str}`")
def when_i_set_the_test_contract_expiration_date_to(context, date_str):
    date_str = process_template_vars(context, date_str)
    resp = requests.put(
        "https://contracts.staging.canonical.com/v1/contract-items",
        auth=(
            context.pro_config.contract_staging_service_account_username,
            context.pro_config.contract_staging_service_account_password,
        ),
        json=[
            {
                # Do not delete any of these fields. They are all required.
                "contractID": "cAOEJ9Vymiwr47-HZ3FnYR_YDCPILpaTfdloKpojyNPE",
                "id": 39398,
                "effectiveFrom": "2024-01-01T00:00:00Z",
                "effectiveTo": date_str,
                "metric": "units",
                "reason": "contract_created",
                "value": 1000,
            }
        ],
    )
    logging.debug(resp.json())
    assert resp.status_code == 200



from collections import defaultdict

import mock
import pytest

from uaclient import livepatch
from uaclient.api.u.pro.security.status.reboot_required.v1 import RebootStatus
from uaclient.entitlements.entitlement_status import (
    ApplicationStatus,
    ContractStatus,
)
from uaclient.security_status import (
    UpdateStatus,
    filter_updates,
    get_livepatch_fixed_cves,
    get_origin_for_installed_package,
    get_ua_info,
    get_update_status,
    security_status_dict,
)
from uaclient.system import KernelInfo
from uaclient.tests.test_apt import mock_origin, mock_package, mock_version

M_PATH = "uaclient.security_status."


class FakeDepCache:
    def get_candidate_ver(self, package: mock.MagicMock):
        try:
            return max(package.version_list)
        except ValueError:
            return None


MOCK_ORIGINS = {
    "now": mock_origin("now", "now", "", ""),
    "third-party": mock_origin("main", "", "other", "some.other.site"),
    "infra": mock_origin(
        "main", "example-infra-security", "UbuntuESM", "esm.ubuntu.com"
    ),
    "apps": mock_origin(
        "main", "example-apps-security", "UbuntuESMApps", "esm.ubuntu.com"
    ),
    "standard-security": mock_origin(
        "main", "example-security", "Ubuntu", "security.ubuntu.com"
    ),
    "archive_main": mock_origin(
        "main", "example-updates", "Ubuntu", "archive.ubuntu.com"
    ),
    "archive_universe": mock_origin(
        "universe", "example-updates", "Ubuntu", "archive.ubuntu.com"
    ),
    "archive_backports": mock_origin(
        "universe", "example-backports", "Ubuntu", "archive.ubuntu.com"
    ),
}

ORIGIN_TO_SERVICE_MOCK = {
    ("UbuntuESM", "example-infra-security"): "esm-infra",
    ("Ubuntu", "example-security"): "standard-security",
    ("UbuntuESMApps", "example-apps-security"): "esm-apps",
}


class TestSecurityStatus:
    @pytest.mark.parametrize(
        "service_name,ua_info,version,expected_result",
        (
            (
                "standard-security",
                {},
                mock_version("1.0"),
                UpdateStatus.AVAILABLE.value,
            ),
            (
                "standard-security",
                {},
                mock_version("0.8"),
                UpdateStatus.AVAILABLE_NOT_PREFERRED.value,
            ),
            (
                "esm",
                {"attached": True, "enabled_services": ["esm"]},
                mock_version("1.0"),
                UpdateStatus.AVAILABLE.value,
            ),
            (
                "esm",
                {"attached": True, "enabled_services": ["esm"]},
                mock_version("1.1"),
                UpdateStatus.AVAILABLE_NOT_PREFERRED.value,
            ),
            (
                "esm",
                {"attached": False},
                mock_version("1.0"),
                UpdateStatus.UNATTACHED.value,
            ),
            (
                "esm",
                {
                    "attached": True,
                    "enabled_services": ["not-esm"],
                    "entitled_services": ["esm"],
                },
                mock_version("1.0"),
                UpdateStatus.NOT_ENABLED.value,
            ),
            (
                "esm",
                {
                    "attached": True,
                    "enabled_services": ["not-esm"],
                    "entitled_services": ["not-esm-again"],
                },
                mock_version("1.0"),
                UpdateStatus.UNAVAILABLE.value,
            ),
        ),
    )
    @mock.patch(M_PATH + "get_pkg_candidate_version", return_value="1.0")
    def test_get_update_status(
        self,
        m_pkg_candidate_version,
        service_name,
        ua_info,
        version,
        expected_result,
    ):
        assert (
            get_update_status(service_name, ua_info, version)
            == expected_result
        )

    @pytest.mark.parametrize("is_attached", (True, False))
    @mock.patch(M_PATH + "_is_attached")
    @mock.patch(M_PATH + "ESMInfraEntitlement")
    @mock.patch(M_PATH + "ESMAppsEntitlement")
    def test_get_ua_info(
        self, m_apps, m_infra, m_attached, is_attached, FakeConfig
    ):
        m_attached.return_value = mock.MagicMock(is_attached=is_attached)

        m_infra.return_value = mock.MagicMock(
            contract_status=mock.MagicMock(
                return_value=ContractStatus.ENTITLED
            ),
            application_status=mock.MagicMock(
                return_value=(ApplicationStatus.ENABLED, 0)
            ),
        )
        m_apps.return_value = mock.MagicMock(
            contract_status=mock.MagicMock(
                return_value=ContractStatus.ENTITLED
            ),
            application_status=mock.MagicMock(
                return_value=(ApplicationStatus.DISABLED, 0)
            ),
        )

        cfg = FakeConfig()
        result = get_ua_info(cfg)

        if is_attached:
            assert result == {
                "attached": True,
                "enabled_services": ["esm-infra"],
                "entitled_services": ["esm-apps", "esm-infra"],
            }
            assert m_infra.call_args_list == [mock.call(cfg)]
            assert m_apps.call_args_list == [mock.call(cfg)]
        else:
            assert result == {
                "attached": False,
                "enabled_services": [],
                "entitled_services": [],
            }
            assert m_infra.call_args_list == []
            assert m_apps.call_args_list == []

    @pytest.mark.parametrize(
        "installed_version,other_versions,expected_output",
        (
            (mock_version("1.0", [MOCK_ORIGINS["now"]]), [], "unknown"),
            (
                mock_version("2.0", [MOCK_ORIGINS["now"]]),
                [mock_version("1.0", [MOCK_ORIGINS["archive_main"]])],
                "unknown",
            ),
            (
                mock_version("2.0", [MOCK_ORIGINS["now"]]),
                [mock_version("3.0", [MOCK_ORIGINS["archive_main"]])],
                "main",
            ),
            (
                mock_version(
                    "1.0", [MOCK_ORIGINS["infra"], MOCK_ORIGINS["now"]]
                ),
                [],
                "esm-infra",
            ),
            (
                mock_version(
                    "1.0", [MOCK_ORIGINS["apps"], MOCK_ORIGINS["now"]]
                ),
                [],
                "esm-apps",
            ),
            (
                mock_version(
                    "1.0", [MOCK_ORIGINS["archive_main"], MOCK_ORIGINS["now"]]
                ),
                [],
                "main",
            ),
            (
                mock_version(
                    "1.0",
                    [MOCK_ORIGINS["archive_universe"], MOCK_ORIGINS["now"]],
                ),
                [],
                "universe",
            ),
            (
                mock_version(
                    "1.0", [MOCK_ORIGINS["third-party"], MOCK_ORIGINS["now"]]
                ),
                [],
                "third-party",
            ),
        ),
    )
    def test_get_origin_for_installed_package(
        self, installed_version, other_versions, expected_output
    ):
        package_mock = mock_package(
            "example", installed_version, other_versions
        )
        with mock.patch(
            M_PATH + "get_origin_information_to_service_map",
            return_value=ORIGIN_TO_SERVICE_MOCK,
        ):
            assert expected_output == get_origin_for_installed_package(
                package_mock, FakeDepCache()
            )

    def test_get_origin_for_installed_package_without_candidate(self):
        """Packages without candidates are unknown."""
        package_mock = mock_package(
            "example", mock_version("1.0", [MOCK_ORIGINS["now"]]), []
        )
        # no candidate will be returned
        package_mock.version_list = []

        fake_dep_cache = FakeDepCache()
        assert fake_dep_cache.get_candidate_ver(package_mock) is None

        with mock.patch(
            M_PATH + "get_origin_information_to_service_map",
            return_value=ORIGIN_TO_SERVICE_MOCK,
        ):
            assert "unknown" == get_origin_for_installed_package(
                package_mock, fake_dep_cache
            )

    @mock.patch(
        "uaclient.security_status.get_esm_apt_pkg_cache", return_value={}
    )
    def test_filter_updates(self, _m_get_esm_apt_pkgcache):
        expected_return = defaultdict(
            list,
            {
                "esm-infra": [
                    (
                        mock_version("2.0", [MOCK_ORIGINS["infra"]]),
                        "esm.ubuntu.com",
                    )
                ],
                "standard-security": [
                    (
                        mock_version(
                            "2.0", [MOCK_ORIGINS["standard-security"]]
                        ),
                        "security.ubuntu.com",
                    ),
                    (
                        mock_version(
                            "2.1", [MOCK_ORIGINS["standard-security"]]
                        ),
                        "security.ubuntu.com",
                    ),
                ],
                "esm-apps": [
                    (
                        mock_version("3.0", [MOCK_ORIGINS["apps"]]),
                        "esm.ubuntu.com",
                    )
                ],
                "standard-updates": [
                    (
                        mock_version("2.0", [MOCK_ORIGINS["archive_main"]]),
                        "archive.ubuntu.com",
                    )
                ],
            },
        )
        package_list = [
            mock_package(name="not-installed"),
            mock_package(
                name="there-is-no-update",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
            ),
            mock_package(
                name="latest-is-installed",
                installed_version=mock_version(
                    "2.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["infra"]]
                ),
                other_versions=[mock_version("1.0", [MOCK_ORIGINS["infra"]])],
            ),
            mock_package(
                name="infra-update-available",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[expected_return["esm-infra"][0][0]],
            ),
            mock_package(
                name="security-update-available",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[
                    expected_return["standard-security"][0][0],
                ],
            ),
            mock_package(
                name="not-a-security-update",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[expected_return["standard-updates"][0][0]],
            ),
            mock_package(
                name="more-than-one-update",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[
                    expected_return["standard-security"][1][0],
                    expected_return["esm-apps"][0][0],
                ],
            ),
            mock_package(
                name="upgrade-is-backports-boo",
                installed_version=mock_version(
                    "1.0",
                    [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_universe"]],
                ),
                other_versions=[
                    mock_version("2.0", [MOCK_ORIGINS["archive_backports"]])
                ],
            ),
        ]
        with mock.patch(
            M_PATH + "get_origin_information_to_service_map",
            return_value=ORIGIN_TO_SERVICE_MOCK,
        ):
            filtered_versions = filter_updates(package_list)
            assert expected_return == filtered_versions
            assert (
                filtered_versions["esm-infra"][0][0].parent_pkg.name
                == "infra-update-available"
            )
            assert (
                filtered_versions["standard-security"][0][0].parent_pkg.name
                == "security-update-available"
            )
            assert (
                filtered_versions["standard-security"][1][0].parent_pkg.name
                == "more-than-one-update"
            )
            assert (
                filtered_versions["esm-apps"][0][0].parent_pkg.name
                == "more-than-one-update"
            )
            assert (
                filtered_versions["standard-updates"][0][0].parent_pkg.name
                == "not-a-security-update"
            )

    @mock.patch("uaclient.security_status.get_esm_apt_pkg_cache")
    def test_filter_updates_when_esm_disabled(self, m_esm_cache):
        expected_return = defaultdict(
            list,
            {
                "esm-infra": [
                    (
                        mock_version("2.0", [MOCK_ORIGINS["infra"]]),
                        "esm.ubuntu.com",
                    )
                ],
                "standard-security": [
                    (
                        mock_version(
                            "2.0", [MOCK_ORIGINS["standard-security"]]
                        ),
                        "security.ubuntu.com",
                    ),
                    (
                        mock_version(
                            "2.1", [MOCK_ORIGINS["standard-security"]]
                        ),
                        "security.ubuntu.com",
                    ),
                ],
                "esm-apps": [
                    (
                        mock_version("3.0", [MOCK_ORIGINS["apps"]]),
                        "esm.ubuntu.com",
                    )
                ],
                "standard-updates": [
                    (
                        mock_version("2.0", [MOCK_ORIGINS["archive_main"]]),
                        "archive.ubuntu.com",
                    )
                ],
            },
        )
        esm_package_list = {
            "infra-update-available": mock_package(
                name="infra-update-available",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[expected_return["esm-infra"][0][0]],
            ),
            "update-multiple-caches": mock_package(
                name="update-multiple-caches",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[
                    expected_return["esm-apps"][0][0],
                ],
            ),
        }

        package_list = [
            mock_package(name="not-installed"),
            mock_package(
                name="latest-is-installed",
                installed_version=mock_version(
                    "2.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["infra"]]
                ),
            ),
            mock_package(
                name="infra-update-available",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
            ),
            mock_package(
                name="there-is-no-update",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
            ),
            mock_package(
                name="security-update-available",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[
                    expected_return["standard-security"][0][0],
                ],
            ),
            mock_package(
                name="not-a-security-update",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[expected_return["standard-updates"][0][0]],
            ),
            mock_package(
                name="update-multiple-caches",
                installed_version=mock_version(
                    "1.0", [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_main"]]
                ),
                other_versions=[
                    expected_return["standard-security"][1][0],
                ],
            ),
            mock_package(
                name="upgrade-is-backports-boo",
                installed_version=mock_version(
                    "1.0",
                    [MOCK_ORIGINS["now"], MOCK_ORIGINS["archive_universe"]],
                ),
                other_versions=[
                    mock_version("2.0", [MOCK_ORIGINS["archive_backports"]])
                ],
            ),
        ]

        m_esm_cache.return_value = esm_package_list
        with mock.patch(
            M_PATH + "get_origin_information_to_service_map",
            return_value=ORIGIN_TO_SERVICE_MOCK,
        ):
            filtered_versions = filter_updates(package_list)
            assert expected_return == filtered_versions
            assert (
                filtered_versions["esm-infra"][0][0].parent_pkg.name
                == "infra-update-available"
            )
            assert (
                filtered_versions["standard-security"][0][0].parent_pkg.name
                == "security-update-available"
            )
            assert (
                filtered_versions["standard-security"][1][0].parent_pkg.name
                == "update-multiple-caches"
            )
            assert (
                filtered_versions["esm-apps"][0][0].parent_pkg.name
                == "update-multiple-caches"
            )
            assert (
                filtered_versions["standard-updates"][0][0].parent_pkg.name
                == "not-a-security-update"
            )

    @mock.patch(M_PATH + "apt_pkg.DepCache", return_value=FakeDepCache())
    @mock.patch(M_PATH + "_reboot_required")
    @mock.patch(M_PATH + "get_livepatch_fixed_cves", return_value=[])
    @mock.patch(
        M_PATH + "_is_attached", return_value=mock.MagicMock(is_attached=False)
    )
    @mock.patch(
        M_PATH + "get_origin_for_installed_package", return_value="main"
    )
    @mock.patch(M_PATH + "filter_updates")
    @mock.patch(M_PATH + "get_apt_pkg_cache")
    @mock.patch(M_PATH + "get_pkg_candidate_version", return_value=None)
    def test_security_status_dict(
        self,
        m_pkg_candidate_version,
        m_cache,
        m_filter_sec_updates,
        _m_get_origin,
        _m_status,
        _m_livepatch_cves,
        m_reboot_status,
        _m_dep_cache,
        FakeConfig,
    ):
        """Make sure the output format matches the expected JSON"""
        cfg = FakeConfig()
        m_version = mock_version("1.0", size=123456)
        m_package = mock_package("example_package", m_version)

        # Different package with version that is not candidate
        m_version_2 = mock_version("2.0", size=123456)
        m_package_2 = mock_package(
            "example_pkg_diff_candidate_version", m_version_2
        )

        m_package_list = [m_package] * 10
        m_package_list.append(m_package_2)
        m_cache.return_value.packages = m_package_list

        m_pkg_candidate_version.return_value = "1.0"

        m_filter_sec_updates.return_value = {
            "esm-infra": [
                (m_version, "some.url.for.esm"),
                (m_version, "some.url.for.esm"),
            ],
            "esm-apps": [],
            "standard-security": [
                (m_version_2, "security.ubuntu.com"),
            ],
        }
        m_reboot_status.return_value = mock.MagicMock(
            reboot_required=RebootStatus.REBOOT_NOT_REQUIRED.value
        )

        expected_output = {
            "_schema_version": "0.1",
            "packages": [
                {
                    "package": "example_package",
                    "version": "1.0",
                    "service_name": "esm-infra",
                    "status": "pending_attach",
                    "origin": "some.url.for.esm",
                    "download_size": 123456,
                },
                {
                    "package": "example_package",
                    "version": "1.0",
                    "service_name": "esm-infra",
                    "status": "pending_attach",
                    "origin": "some.url.for.esm",
                    "download_size": 123456,
                },
                {
                    "package": "example_pkg_diff_candidate_version",
                    "version": "2.0",
                    "service_name": "standard-security",
                    "status": "upgrade_available_not_preferred",
                    "origin": "security.ubuntu.com",
                    "download_size": 123456,
                },
            ],
            "summary": {
                "ua": {
                    "attached": False,
                    "enabled_services": [],
                    "entitled_services": [],
                },
                "num_installed_packages": 11,
                "num_main_packages": 11,
                "num_restricted_packages": 0,
                "num_universe_packages": 0,
                "num_multiverse_packages": 0,
                "num_third_party_packages": 0,
                "num_unknown_packages": 0,
                "num_esm_infra_packages": 0,
                "num_esm_apps_packages": 0,
                "num_esm_infra_updates": 2,
                "num_esm_apps_updates": 0,
                "num_standard_security_updates": 1,
                "reboot_required": "no",
            },
            "livepatch": {"fixed_cves": []},
        }

        result = security_status_dict(cfg)
        assert expected_output["_schema_version"] == result["_schema_version"]
        assert expected_output["livepatch"] == result["livepatch"]
        assert expected_output["summary"] == result["summary"]
        assert expected_output["packages"] == sorted(
            result["packages"], key=lambda x: x["version"]
        )


@mock.patch(M_PATH + "livepatch.status")
@mock.patch(M_PATH + "get_kernel_info")
class TestGetLivepatchFixedCVEs:
    def test_livepatch_status_none(self, _m_kernel_info, m_livepatch_status):
        m_livepatch_status.return_value = None
        assert [] == get_livepatch_fixed_cves()

    def test_cant_get_kernel_info(self, m_kernel_info, m_livepatch_status):
        m_livepatch_status.return_value = livepatch.LivepatchStatusStatus(
            kernel="installed-kernel-generic",
            livepatch=livepatch.LivepatchPatchStatus(
                state="nothing-to-apply", fixes=None, version=None
            ),
            supported=None,
        )

        m_kernel_info.return_value = KernelInfo(
            uname_machine_arch="",
            uname_release="",
            proc_version_signature_version=None,
            build_date=None,
            major=None,
            minor=None,
            patch=None,
            abi=None,
            flavor=None,
        )

        assert [] == get_livepatch_fixed_cves()

    def test_livepatch_no_fixes(self, m_kernel_info, m_livepatch_status):
        m_kernel_info.return_value.proc_version_signature_version = (
            "installed-kernel-generic"
        )
        m_livepatch_status.return_value = livepatch.LivepatchStatusStatus(
            kernel="installed-kernel-generic",
            livepatch=livepatch.LivepatchPatchStatus(
                state="nothing-to-apply", fixes=None, version=None
            ),
            supported=None,
        )

        assert [] == get_livepatch_fixed_cves()

    def test_livepatch_has_fixes(self, m_kernel_info, m_livepatch_status):
        m_kernel_info.return_value.proc_version_signature_version = (
            "installed-kernel-generic"
        )
        m_livepatch_status.return_value = livepatch.LivepatchStatusStatus(
            kernel="installed-kernel-generic",
            livepatch=livepatch.LivepatchPatchStatus(
                state="applied",
                fixes=[
                    livepatch.LivepatchPatchFixStatus(
                        name="cve-example", patched=True
                    )
                ],
                version=None,
            ),
            supported=None,
        )

        assert [
            {"name": "cve-example", "patched": True}
        ] == get_livepatch_fixed_cves()



"""
This is called in postinst when upgrading from a version <27.14
This removes the "ua_config" sub-document from uaclient.conf and moves it to
it's own json file in /var/lib/ubuntu-advantage/user-config.json.
The fields under "ua_config" are intended to be set/unset by the user via the
`pro config set` command, so we don't want them covered as a "conffile".
It writes back a uaclient.conf that will match the new default uaclient.conf
as long as no defaults have changed.
We print warning messages if errors occur with instructions to help the user
resolve them.
"""

import json
import os
import sys

from uaclient import defaults, messages
from uaclient.yaml import safe_dump, safe_load

UACLIENT_CONF_BACKUP_PATH = (
    "/etc/ubuntu-advantage/uaclient.conf.preinst-backup"
)
UACLIENT_CONF_MIGRATED_TEMP_PATH = (
    "/etc/ubuntu-advantage/uaclient.conf.preinst-backup-migrated-temp"
)

USER_CONFIG_DEFAULTS = {
    "http_proxy": None,
    "https_proxy": None,
    "apt_http_proxy": None,
    "apt_https_proxy": None,
    "ua_apt_http_proxy": None,
    "ua_apt_https_proxy": None,
    "global_apt_http_proxy": None,
    "global_apt_https_proxy": None,
    "update_messaging_timer": 21600,
    "metering_timer": 14400,
    "apt_news": True,
    "apt_news_url": "https://motd.ubuntu.com/aptnews.json",
}


def load_pre_upgrade_conf():
    # Step 1: Load pre-upgrade uaclient.conf backed up in preinst
    try:
        with open(UACLIENT_CONF_BACKUP_PATH, "r") as uaclient_conf_file:
            old_uaclient_conf = safe_load(uaclient_conf_file)
    except Exception:
        print(
            messages.USER_CONFIG_MIGRATION_WARNING_UACLIENT_CONF_LOAD,
            file=sys.stderr,
        )
        return None
    return old_uaclient_conf


def create_new_user_config_file(old_uaclient_conf):
    # Step 2: Create new user_config.json
    old_user_config = old_uaclient_conf.get("ua_config", {})
    if not isinstance(old_user_config, dict):
        # invalid and could not have been working, just ignore by treating
        # as empty
        old_user_config = {}

    new_user_config = {}
    # only keep a setting if it was changed
    for field in USER_CONFIG_DEFAULTS.keys():
        old_val = old_user_config.get(field)
        if old_val is not None and old_val != USER_CONFIG_DEFAULTS.get(field):
            new_user_config[field] = old_val

    try:
        with open(
            defaults.DEFAULT_USER_CONFIG_JSON_FILE, "w"
        ) as user_config_file:
            json.dump(new_user_config, user_config_file)
    except Exception:
        if len(new_user_config) > 0:
            print(
                messages.USER_CONFIG_MIGRATION_WARNING_NEW_USER_CONFIG_WRITE,
                file=sys.stderr,
            )
            for field in sorted(new_user_config.keys()):
                print(
                    "           pro config set {}={}".format(
                        field, new_user_config[field]
                    ),
                    file=sys.stderr,
                )


def create_new_uaclient_conffile(old_uaclient_conf):
    # Step 3: Create new uaclient.conf
    # The goal here is to end up with a minimal diff compared to the default
    # uaclient.conf.
    # If nothing has changed in any of the uaclient.conf fields, then the
    # result should exactly match the new default uaclient.conf.
    new_uaclient_conf = {
        "contract_url": old_uaclient_conf.get(
            "contract_url", defaults.BASE_CONTRACT_URL
        ),
        "log_level": old_uaclient_conf.get("log_level", "debug"),
    }

    # these are only present if they were added by the user
    for field in ("features", "settings_overrides"):
        if field in old_uaclient_conf:
            new_uaclient_conf[field] = old_uaclient_conf.get(field)

    # the rest only if they're different from defaults
    for field in (
        "data_dir",
        "log_file",
        "security_url",
    ):
        old_val = old_uaclient_conf.get(field)
        if old_val is not None and old_val != defaults.CONFIG_DEFAULTS.get(
            field
        ):
            new_uaclient_conf[field] = old_val

    try:
        print(messages.USER_CONFIG_MIGRATION_MIGRATING, file=sys.stderr)
        new_uaclient_conf_str = safe_dump(
            new_uaclient_conf,
            default_flow_style=False,
        )
        # don't open until after successful yaml serialization
        # this way, if there is an error in yaml serialization we won't
        # accidentally truncate the file
        with open(UACLIENT_CONF_MIGRATED_TEMP_PATH, "w") as uaclient_conf_file:
            uaclient_conf_file.write(new_uaclient_conf_str)

        # write to a temp file and rename atomically to avoid partial writes
        os.rename(UACLIENT_CONF_MIGRATED_TEMP_PATH, UACLIENT_CONF_BACKUP_PATH)
    except Exception:
        print(
            messages.USER_CONFIG_MIGRATION_WARNING_NEW_UACLIENT_CONF_WRITE,
            file=sys.stderr,
        )
        for field in sorted(new_uaclient_conf.keys()):
            print(
                "           {}: {}".format(
                    field, repr(new_uaclient_conf[field])
                ),
                file=sys.stderr,
            )


def main():
    old_uaclient_conf = load_pre_upgrade_conf()
    if old_uaclient_conf is None:
        # something went wrong, we can't continue migration
        return
    create_new_user_config_file(old_uaclient_conf)
    create_new_uaclient_conffile(old_uaclient_conf)


if __name__ == "__main__":
    main()



import datetime

import mock
import pytest

import uaclient.api.u.unattended_upgrades.status.v1 as api
from uaclient.messages import (
    UNATTENDED_UPGRADES_CFG_LIST_VALUE_EMPTY,
    UNATTENDED_UPGRADES_CFG_VALUE_TURNED_OFF,
    UNATTENDED_UPGRADES_SYSTEMD_JOB_DISABLED,
    UNATTENDED_UPGRADES_UNINSTALLED,
)

M_PATH = "uaclient.api.u.unattended_upgrades.status.v1"


class TestUnattendedUpgradesGetAptDailyJob:
    @pytest.mark.parametrize(
        [
            "systemd_unit_active_return",
            "expected_return",
        ],
        (
            ((True, True), True),
            ((False, True), False),
            ((True, False), False),
            ((False, False), False),
        ),
    )
    @mock.patch(M_PATH + ".system.is_systemd_unit_active")
    def test_get_apt_daily_job_status(
        self,
        m_is_systemd_unit_active,
        systemd_unit_active_return,
        expected_return,
    ):
        m_is_systemd_unit_active.side_effect = systemd_unit_active_return
        assert expected_return is api._get_apt_daily_job_status()


class TestIsUnattendedUpgradesRunning:
    @pytest.mark.parametrize(
        "apt_timer_enabled,unattended_upgrades_cfg,expected_ret,expected_msg",
        (
            (False, {}, False, UNATTENDED_UPGRADES_SYSTEMD_JOB_DISABLED),
            (
                True,
                {
                    "APT::Periodic::Enable": "1",
                    "APT::Periodic::Update-Package-Lists": "1",
                    "APT::Periodic::Unattended-Upgrade": "1",
                    "Unattended-Upgrade::Allowed-Origins": [],
                },
                False,
                UNATTENDED_UPGRADES_CFG_LIST_VALUE_EMPTY.format(
                    cfg_name="Unattended-Upgrade::Allowed-Origins"
                ),
            ),
            (
                True,
                {
                    "APT::Periodic::Enable": "0",
                    "APT::Periodic::Update-Package-Lists": "0",
                    "APT::Periodic::Unattended-Upgrade": "0",
                    "Unattended-Upgrade::Allowed-Origins": ["foo"],
                },
                False,
                UNATTENDED_UPGRADES_CFG_VALUE_TURNED_OFF.format(
                    cfg_name="APT::Periodic::Enable"
                ),
            ),
            (
                True,
                {
                    "APT::Periodic::Enable": "1",
                    "APT::Periodic::Update-Package-Lists": "0",
                    "APT::Periodic::Unattended-Upgrade": "1",
                    "Unattended-Upgrade::Allowed-Origins": ["foo"],
                },
                False,
                UNATTENDED_UPGRADES_CFG_VALUE_TURNED_OFF.format(
                    cfg_name="APT::Periodic::Update-Package-Lists"
                ),
            ),
            (
                True,
                {
                    "APT::Periodic::Enable": "1",
                    "APT::Periodic::Update-Package-Lists": None,
                    "APT::Periodic::Unattended-Upgrade": "1",
                    "Unattended-Upgrade::Allowed-Origins": ["foo"],
                },
                False,
                UNATTENDED_UPGRADES_CFG_VALUE_TURNED_OFF.format(
                    cfg_name="APT::Periodic::Update-Package-Lists"
                ),
            ),
            (
                True,
                {
                    "APT::Periodic::Enable": "1",
                    "APT::Periodic::Update-Package-Lists": "1",
                    "APT::Periodic::Unattended-Upgrade": "1",
                    "Unattended-Upgrade::Allowed-Origins": ["foo"],
                },
                True,
                None,
            ),
            (
                True,
                {
                    "APT::Periodic::Enable": "1",
                    "APT::Periodic::Update-Package-Lists": "1",
                    "APT::Periodic::Unattended-Upgrade": "1",
                    "Unattended-Upgrade::Allowed-Origins": ["foo"],
                    "Unattended-Upgrade::OnlyOnACPower": "0",
                },
                True,
                None,
            ),
        ),
    )
    def test_is_unattended_upgrades_running(
        self,
        apt_timer_enabled,
        unattended_upgrades_cfg,
        expected_ret,
        expected_msg,
    ):
        assert (
            expected_ret,
            expected_msg,
        ) == api._is_unattended_upgrades_running(
            apt_timer_enabled, unattended_upgrades_cfg
        )


class TestUnattendedUpgradesLastRun:
    @mock.patch("os.path.getctime")
    def test_unattended_upgrades_last_run_when_file_not_present(
        self,
        m_getctime,
    ):
        m_getctime.side_effect = FileNotFoundError()
        assert None is api._get_unattended_upgrades_last_run()


class TestUnattendedUpgradesStatusV1:
    @mock.patch("uaclient.apt.is_installed", return_value=True)
    @mock.patch(M_PATH + ".get_apt_config_keys")
    @mock.patch(M_PATH + ".get_apt_config_values")
    @mock.patch(M_PATH + "._is_unattended_upgrades_running")
    @mock.patch(M_PATH + "._get_apt_daily_job_status")
    @mock.patch(M_PATH + "._get_unattended_upgrades_last_run")
    def test_unattended_upgrades_status_v1(
        self,
        m_last_run,
        m_apt_job_status,
        m_is_running,
        m_apt_cfg_values,
        m_apt_cfg_keys,
        _m_apt_is_installed,
        FakeConfig,
    ):
        expected_datetime = datetime.datetime(2023, 2, 23, 15, 0, 0, 102490)

        m_is_running.return_value = (True, "")
        m_apt_job_status.return_value = True
        m_last_run.return_value = expected_datetime
        m_apt_cfg_values.return_value = {
            "APT::Periodic::Enable": "",
            "APT::Periodic::Update-Package-Lists": "1",
            "APT::Periodic::Unattended-Upgrade": "1",
            "Unattended-Upgrade::Allowed-Origins": ["test"],
            "Unattended-Upgrade::Mail": "mail",
        }
        m_apt_cfg_keys.return_value = ["Unattended-Upgrade::Mail"]

        actual_return = api._status(FakeConfig())
        assert True is actual_return.apt_periodic_job_enabled
        assert True is actual_return.systemd_apt_timer_enabled
        assert 1 == actual_return.package_lists_refresh_frequency_days
        assert 1 == actual_return.unattended_upgrades_frequency_days
        assert ["test"] == actual_return.unattended_upgrades_allowed_origins
        assert True is actual_return.unattended_upgrades_running
        assert expected_datetime == actual_return.unattended_upgrades_last_run
        assert None is actual_return.unattended_upgrades_disabled_reason
        assert (
            "1"
            == actual_return.meta["raw_config"][
                "APT::Periodic::Unattended-Upgrade"
            ]
        )
        assert "1" == actual_return.meta["raw_config"]["APT::Periodic::Enable"]
        assert (
            "1"
            == actual_return.meta["raw_config"][
                "APT::Periodic::Update-Package-Lists"
            ]
        )
        assert ["test"] == actual_return.meta["raw_config"][
            "Unattended-Upgrade::Allowed-Origins"
        ]
        assert (
            "mail"
            == actual_return.meta["raw_config"]["Unattended-Upgrade::Mail"]
        )

        assert m_is_running.call_count == 1
        assert m_apt_job_status.call_count == 1
        assert m_apt_cfg_values.call_count == 1
        assert m_last_run.call_count == 1
        assert _m_apt_is_installed.call_count == 1

    @mock.patch("uaclient.apt.is_installed", return_value=False)
    def test_unattended_upgrades_status_v1_when_pkg_not_installed(
        self,
        _m_apt_is_installed,
        FakeConfig,
    ):
        actual_return = api._status(FakeConfig())
        assert False is actual_return.apt_periodic_job_enabled
        assert False is actual_return.systemd_apt_timer_enabled
        assert 0 == actual_return.package_lists_refresh_frequency_days
        assert 0 == actual_return.unattended_upgrades_frequency_days
        assert [] == actual_return.unattended_upgrades_allowed_origins
        assert False is actual_return.unattended_upgrades_running
        assert None is actual_return.unattended_upgrades_last_run
        assert (
            api.UnattendedUpgradesDisabledReason(
                msg=UNATTENDED_UPGRADES_UNINSTALLED.msg,
                code=UNATTENDED_UPGRADES_UNINSTALLED.name,
            )
            == actual_return.unattended_upgrades_disabled_reason
        )

        assert _m_apt_is_installed.call_count == 1



import datetime
import logging
import sys
from typing import NamedTuple

from behave import given, when
from pycloudlib.instance import BaseInstance  # type: ignore

from features.steps.packages import when_i_apt_install, when_i_apt_update
from features.steps.shell import when_i_run_command
from features.steps.ubuntu_advantage_tools import when_i_install_uat
from features.util import (
    BUILDER_NAME_PREFIX,
    SUT,
    InstallationSource,
    build_debs,
    get_debs_for_series,
)

MachineTuple = NamedTuple(
    "MachineTuple",
    [
        ("series", str),
        ("instance", BaseInstance),
        ("machine_type", str),
        ("cloud", str),
    ],
)
SnapshotTuple = NamedTuple(
    "SnapshotTuple",
    [("series", str), ("name", str), ("machine_type", str), ("cloud", str)],
)


@when(
    "I launch a `{series}` `{machine_type}` machine named `{machine_name}` from the snapshot of `{snapshot_name}`"  # noqa: E501
)
@given("a `{series}` `{machine_type}` machine")
@given("a `{series}` `{machine_type}` machine named `{machine_name}`")
@given(
    "a `{series}` `{machine_type}` machine named `{machine_name}` with ingress ports `{ports}`"  # noqa: E501
)
def given_a_machine(
    context,
    series,
    machine_type,
    machine_name=SUT,
    snapshot_name=None,
    user_data=None,
    ports=None,
    cleanup=True,
):
    cloud = machine_type.split(".")[0]
    context.pro_config.clouds.get(cloud).manage_ssh_key()

    time_suffix = datetime.datetime.now().strftime("%m%d-%H%M%S%f")
    instance_name = "pro-{series}-{machine_name}-{time_suffix}".format(
        series=series,
        machine_name=machine_name,
        time_suffix=time_suffix,
    )

    inbound_ports = ports.split(",") if ports is not None else None

    is_pro = "pro" in machine_type
    pro_user_data = (
        "bootcmd:\n"
        """  - "cloud-init-per once disable-auto-attach printf '\\nfeatures: {disable_auto_attach: true}\\n' >> /etc/ubuntu-advantage/uaclient.conf"\n"""  # noqa: E501
    )
    user_data_to_use = None
    if is_pro or user_data is not None:
        user_data_to_use = "#cloud-config\n"
        if is_pro:
            user_data_to_use += pro_user_data
        if user_data is not None:
            user_data_to_use += user_data

    if snapshot_name and snapshot_name in context.snapshots:
        image_name = context.snapshots[snapshot_name].name
    else:
        image_name = None

    instance = context.pro_config.clouds.get(cloud).launch(
        series=series,
        machine_type=machine_type,
        instance_name=instance_name,
        ephemeral=context.pro_config.ephemeral_instance,
        image_name=image_name,
        inbound_ports=inbound_ports,
        user_data=user_data_to_use,
    )

    context.machines[machine_name] = MachineTuple(
        series=series,
        instance=instance,
        machine_type=machine_type,
        cloud=cloud,
    )

    if series == "xenial":
        # Upgrading open-iscsi to esm version on xenial restarts this service
        # This sometimes causes resource errors on github action runners
        when_i_run_command(
            context,
            "systemctl mask iscsid.service",
            "with sudo",
            machine_name=machine_name,
        )

    when_i_run_command(
        context,
        "systemctl mask apt-news.service",
        "with sudo",
        machine_name=machine_name,
    )

    # make sure the machine has up-to-date apt data
    when_i_apt_update(context, machine_name=machine_name)

    # add coverage
    if context.pro_config.collect_coverage:
        when_i_apt_install(
            context, "python3-coverage", machine_name=machine_name
        )

    when_i_run_command(
        context,
        "systemctl unmask apt-news.service",
        "with sudo",
        machine_name=machine_name,
    )

    if cleanup:

        def cleanup_instance():
            if not context.pro_config.destroy_instances:
                logging.info(
                    "--- Leaving instance running: {}".format(
                        context.machines[machine_name].instance.name
                    )
                )
                return
            try:
                machine = context.machines.pop(machine_name)
                machine.instance.delete(wait=False)
            except RuntimeError as e:
                logging.error(
                    "Failed to delete instance: {}\n{}".format(
                        context.machines[machine_name].instance.name, str(e)
                    )
                )

        context.add_cleanup(cleanup_instance)


@when("I take a snapshot of the machine")
def when_i_take_a_snapshot(context, machine_name=SUT, cleanup=True):
    machine_type = context.machines[machine_name].machine_type
    series = context.machines[machine_name].series
    cloud = context.machines[machine_name].cloud
    inst = context.machines[machine_name].instance
    snapshot = context.pro_config.clouds.get(cloud).api.snapshot(inst)

    context.snapshots[machine_name] = SnapshotTuple(
        series=series, name=snapshot, machine_type=machine_type, cloud=cloud
    )

    if cleanup:

        def cleanup_snapshot() -> None:
            try:
                context.pro_config.clouds.get(cloud).api.delete_image(snapshot)
            except RuntimeError as e:
                logging.error(
                    "Failed to delete image: {}\n{}".format(snapshot, str(e))
                )

        context.add_cleanup(cleanup_snapshot)


def _update_distro_info_data(context, machine_name=SUT):
    # There is a problem on Xenial where distro-info-data was incorrectly
    # saying that Xenial was only supported until 2024. The fix was
    # SRUed to Xenial, but now we need to guarantee that we run our tests
    # on the latest version of that package. Additionally, we don't see
    # a problem of always upgrading that package for every release.
    when_i_run_command(
        context,
        "apt install distro-info-data -y",
        "with sudo",
        machine_name=machine_name,
    )
    # And this will kick of the ESM cache setup
    when_i_apt_update(context, machine_name=machine_name)


@given(
    "a `{series}` `{machine_type}` machine with ubuntu-advantage-tools installed"  # noqa: E501
)
def given_a_sut_machine(context, series, machine_type):
    if context.pro_config.install_from == InstallationSource.LOCAL:
        # build right away, this will cache the built debs for later use
        # building early means we catch build errors before investing in
        # launching instances
        build_debs(
            series,
            sbuild_output_to_terminal=context.pro_config.sbuild_output_to_terminal,  # noqa: E501
        )

    if context.pro_config.install_from == InstallationSource.PREBUILT:
        deb_paths = get_debs_for_series(context.pro_config.debs_path, series)
        if not deb_paths:
            logging.error(
                (
                    "UACLIENT_BEHAVE_INSTALL_FROM is set to 'prebuilt', "
                    "but no debs for series {} were found in"
                    "UACLIENT_BEHAVE_DEBS_PATH"
                ).format(series)
            )
            sys.exit(1)

    builder_name = BUILDER_NAME_PREFIX + machine_type + "-" + series

    if context.pro_config.snapshot_strategy:
        if builder_name not in context.snapshots:
            given_a_machine(
                context,
                series,
                machine_type=machine_type,
                machine_name=builder_name,
                cleanup=False,
            )
            _update_distro_info_data(context, machine_name=builder_name)
            when_i_install_uat(context, machine_name=builder_name)
            when_i_take_a_snapshot(
                context,
                machine_name=builder_name,
                cleanup=False,
            )
            context.machines[builder_name].instance.delete(wait=False)
        given_a_machine(
            context,
            series,
            machine_type=machine_type,
            snapshot_name=builder_name,
        )
    else:
        given_a_machine(context, series, machine_type=machine_type)
        when_i_run_command(
            context,
            "systemctl mask apt-news.service",
            "with sudo",
        )
        _update_distro_info_data(context)
        when_i_install_uat(context)

        when_i_run_command(
            context,
            "systemctl unmask apt-news.service",
            "with sudo",
        )

    # trigger GH: #3137 on all machines
    when_i_run_command(
        context,
        "touch /var/lib/dpkg/arch",
        "with sudo",
    )

    logging.info(
        "--- instance ip: {}".format(context.machines[SUT].instance.ip)
    )


@given(
    "a `{series}` `{machine_type}` machine with ubuntu-advantage-tools installed adding this cloud-init user_data:"  # noqa: E501
)
def given_a_sut_machine_with_user_data(context, series, machine_type):
    # doesn't support snapshot strategy because the test depends on
    # custom user data
    given_a_machine(context, series, machine_type, user_data=context.text)
    when_i_install_uat(context)


@when("I reboot the `{machine_name}` machine")
@when("I reboot the machine")
def when_i_reboot_the_machine(context, machine_name=SUT):
    context.machines[machine_name].instance.restart(wait=True)


@when("I update the series in the machine test metadata to `{series}`")
def when_i_update_series(context, series, machine_name=SUT):
    context.machines[machine_name] = MachineTuple(
        series=series,
        machine_type=context.machines[machine_name].machine_type,
        cloud=context.machines[machine_name].cloud,
        instance=context.machines[machine_name].instance,
    )



import logging

import mock
import pytest

from lib.auto_attach import check_cloudinit_userdata_for_ua_info, main
from uaclient import exceptions, messages
from uaclient.api.exceptions import (
    AlreadyAttachedError,
    AutoAttachDisabledError,
)
from uaclient.daemon import AUTO_ATTACH_STATUS_MOTD_FILE


class TestCheckCloudinitUserdataForUAInfo:
    @mock.patch("lib.auto_attach.get_cloudinit_init_stage")
    def test_check_cloudinit_data_returns_false_if_no_cloudinit(
        self, m_get_cloudinit_stage
    ):
        m_get_cloudinit_stage.return_value = None
        assert False is check_cloudinit_userdata_for_ua_info()

    @pytest.mark.parametrize(
        "expected,cloud_cfg",
        (
            (
                False,
                "",
            ),
            (
                False,
                {
                    "runcmd": ["echo test", "echo test2"],
                    "cloud_config_modules": ["ubuntu-advantage", "test"],
                },
            ),
            (
                True,
                {
                    "ubuntu_advantage": {"token": "TOKEN"},
                    "cloud_config_modules": ["ubuntu-advantage", "test"],
                },
            ),
            (
                True,
                {
                    "ubuntu_pro": {"token": "TOKEN"},
                    "cloud_config_modules": ["ubuntu-advantage", "test"],
                },
            ),
            (
                True,
                {
                    "ubuntu-advantage": {"token": "TOKEN"},
                    "cloud_config_modules": ["ubuntu-advantage", "test"],
                },
            ),
        ),
    )
    @mock.patch("lib.auto_attach.get_cloudinit_init_stage")
    def test_check_cloudinit_data(
        self, m_get_cloudinit_stage, expected, cloud_cfg
    ):
        init_mock = mock.MagicMock()
        type(init_mock).cfg = mock.PropertyMock(return_value=cloud_cfg)
        m_get_cloudinit_stage.return_value = init_mock
        assert expected is check_cloudinit_userdata_for_ua_info()


@mock.patch("lib.auto_attach.system.ensure_file_absent")
@mock.patch("lib.auto_attach.system.write_file")
class TestMain:
    @pytest.mark.parametrize(
        "ubuntu_advantage_in_userdata",
        (
            (True,),
            (False,),
        ),
    )
    @mock.patch("lib.auto_attach.check_cloudinit_userdata_for_ua_info")
    @mock.patch("lib.auto_attach.full_auto_attach")
    def test_main(
        self,
        m_api_full_auto_attach,
        m_check_cloudinit,
        m_write_file,
        m_ensure_file_absent,
        ubuntu_advantage_in_userdata,
        FakeConfig,
    ):
        m_check_cloudinit.return_value = ubuntu_advantage_in_userdata
        main(cfg=FakeConfig())

        if not ubuntu_advantage_in_userdata:
            assert [
                mock.call(
                    AUTO_ATTACH_STATUS_MOTD_FILE, messages.AUTO_ATTACH_RUNNING
                )
            ] == m_write_file.call_args_list
            assert 1 == m_api_full_auto_attach.call_count
            assert [
                mock.call(AUTO_ATTACH_STATUS_MOTD_FILE)
            ] == m_ensure_file_absent.call_args_list
        else:
            assert 0 == m_api_full_auto_attach.call_count
            assert 0 == m_write_file.call_count

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @pytest.mark.parametrize(
        "api_side_effect, log_msg, cloud_instance_factory_side_effect",
        [
            (
                AlreadyAttachedError(account_name="test_account"),
                "This machine is already attached to 'test_account'",
                None,
            ),
            (
                AutoAttachDisabledError,
                "Skipping auto-attach. Config disable_auto_attach is set.\n",
                None,
            ),
            (
                None,
                "Error loading the cloud: Unable to determine cloud"
                " platform.\n",
                exceptions.CloudFactoryNoCloudError,
            ),
            (
                None,
                "Error loading the cloud: Auto-attach image support is not"
                " available on this image\n",
                exceptions.CloudFactoryNonViableCloudError,
            ),
            (
                None,
                "Error loading the cloud: Auto-attach image support is not"
                " available on cloud-test\n",
                exceptions.NonAutoAttachImageError(cloud_type="cloud-test"),
            ),
        ],
    )
    @mock.patch("lib.auto_attach.cloud_instance_factory")
    @mock.patch("lib.auto_attach.check_cloudinit_userdata_for_ua_info")
    @mock.patch("lib.auto_attach.full_auto_attach")
    def test_main_handles_errors(
        self,
        m_api_full_auto_attach,
        m_check_cloudinit,
        m_cloud_instance_factory,
        m_write_file,
        m_ensure_file_absent,
        api_side_effect,
        log_msg,
        cloud_instance_factory_side_effect,
        caplog_text,
        fake_machine_token_file,
    ):
        fake_machine_token_file.attached = True
        m_check_cloudinit.return_value = False
        if api_side_effect is not None:
            m_api_full_auto_attach.side_effect = api_side_effect
        if cloud_instance_factory_side_effect is not None:
            m_cloud_instance_factory.side_effect = (
                cloud_instance_factory_side_effect
            )

        main(cfg=None)

        if cloud_instance_factory_side_effect is None:
            assert (
                mock.call(
                    AUTO_ATTACH_STATUS_MOTD_FILE, messages.AUTO_ATTACH_RUNNING
                )
                in m_write_file.call_args_list
            )
        assert m_api_full_auto_attach.call_count == 1
        assert log_msg in caplog_text()
        assert (
            mock.call(AUTO_ATTACH_STATUS_MOTD_FILE)
            in m_ensure_file_absent.call_args_list
        )
